import logging
from collections.abc import Callable, Sequence
from io import BytesIO
from multiprocessing import Pool
from pathlib import Path

import cv2
import h5py
import numpy as np

from idtrackerai import Blob
from idtrackerai.utils import (
    LOGGING_QUEUE,
    Episode,
    IdtrackeraiError,
    setup_logging_queue,
    track,
)


def segment_episode(
    inputs: tuple[Episode, dict],
) -> tuple[list[list[Blob]], Episode, BytesIO | Path]:
    """Gets list of blobs segmented in every frame of the episode of the video
    given by `path` (if the video is splitted in different files) or by
    `episode_start_end_frames` (if the video is given in a single file)

    Parameters
    ----------
    session : <Session object>
        Object collecting all the parameters of the video and paths for saving and loading
    segmentation_thresholds : dict
        Dictionary with the thresholds used for the segmentation: `min_threshold`,
        `max_threshold`, `min_area`, `max_area`
    path : string
        Path to the video file from where to get the VideoCapture (OpenCV) object
    episode_start_end_frames : tuple
        Tuple (starting_frame, ending_frame) indicanting the start and end of the episode
        when the video is given in a single file

    Returns
    -------
    blobs_in_episode : list
        List of `blobs_in_frame` of the episode of the video being segmented
    """
    episode, segmentation_parameters = inputs

    if isinstance(episode.bbox_images, Path):
        # bbox images are saved in disk
        bbox_images_file = episode.bbox_images
    else:
        # bbox images are saved in RAM
        bbox_images_file = BytesIO()

    # Read video for the episode
    cap = cv2.VideoCapture(str(episode.video_path))
    cap.read()  # this somehow initializes the caption and makes following actions less prone to errors, magically

    # Set the video to the starting frame
    success = cap.set(cv2.CAP_PROP_POS_FRAMES, episode.local_start)

    # check where the vide has been really set at
    video_set_at = int(cap.get(cv2.CAP_PROP_POS_FRAMES))

    if video_set_at < episode.local_start or not success:
        # I think this never happens
        logging.error(
            f"OpenCV could not set {episode.video_path} to the starting frame of"
            f" episode {episode.index} (frame {episode.local_start}). Frames from"
            f" {episode.global_start} to {episode.global_end} will be empty."
        )
        return [[] for _ in range(episode.length)], episode, bbox_images_file

    n_error_frames = min(video_set_at - episode.local_start, episode.length)

    if n_error_frames:
        logging.error(
            f"OpenCV could not set video {episode.video_path.name!r} to the starting"
            f" frame of episode {episode.index} (frame {episode.local_start}). Frames"
            f" from {episode.global_start} to"
            f" {episode.global_start + n_error_frames} will be empty."
        )

    blobs_in_episode = [[] for _ in range(n_error_frames)]

    with h5py.File(bbox_images_file, "w") as h5_file:
        for local_frame_number, global_frame_number in zip(
            range(episode.local_start + n_error_frames, episode.local_end),
            range(episode.global_start + n_error_frames, episode.global_end),
        ):
            successfuly_read, frame = cap.read()
            if successfuly_read:
                blobs_in_frame = get_blobs_in_frame(
                    frame, segmentation_parameters, global_frame_number, h5_file
                )
            else:
                logging.error(
                    "OpenCV could not read frame "
                    f"{local_frame_number} of {episode.video_path}"
                )
                blobs_in_frame = []

            # store all the blobs encountered in the episode
            blobs_in_episode.append(blobs_in_frame)

    cap.release()
    return blobs_in_episode, episode, bbox_images_file


def get_blobs_in_frame(
    frame: np.ndarray,
    segmentation_parameters: dict,
    global_frame_number: int,
    bbox_images_file: h5py.File,
) -> list[Blob]:
    """Segments a frame read from `cap` according to the preprocessing parameters
    in `video`. Returns a list `blobs_in_frame` with the Blob objects in the frame
    and the `max_number_of_blobs` found in the video so far. Frames are segmented
    in gray scale.

    Parameters
    ----------
    segmentation_thresholds : dict
        Dictionary with the thresholds used for the segmentation: `min_threshold`,
        `max_threshold`, `min_area`, `max_area`
    global_frame_number : int
        This is the frame number in the whole video. It will be different to the frame_number
        if the video is chuncked.


    Returns
    -------
    list[Blob]
        List of Blob segmented in the current frame
    """

    intensity_ths = segmentation_parameters["intensity_ths"]
    if segmentation_parameters["bkg_model"] is not None and intensity_ths[0] == 0:
        # versions < 5.2.5 used intensity ths with background as (0, value)
        # now we use (value, 255)
        segmentation_parameters["intensity_ths"] = (intensity_ths[1], 255)

    _, contours, frame = process_frame(frame, **segmentation_parameters)

    blobs_in_frame: list[Blob] = []
    for i, contour in enumerate(contours):
        dataset_name = f"{global_frame_number}-{i}"
        bbox_images_file.create_dataset(
            dataset_name, data=get_bbox_image(frame, contour)
        )
        blobs_in_frame.append(Blob(contour, global_frame_number, dataset_name))

    return blobs_in_frame


def process_frame(
    frame: np.ndarray,
    intensity_ths: Sequence[float],
    area_ths: Sequence[float],
    ROI_mask: np.ndarray | None = None,
    bkg_model: np.ndarray | None = None,
) -> tuple[list[int], list[np.ndarray], np.ndarray]:
    # Convert the frame to gray scale
    frame = to_gray_scale(frame)

    if bkg_model is None:
        segmented_frame = cv2.inRange(frame, intensity_ths[0], intensity_ths[1])  # type: ignore
    else:
        segmented_frame = (cv2.absdiff(bkg_model, frame) > intensity_ths[0]).astype(  # type: ignore
            np.uint8, copy=False
        )

    # Applying the mask
    if ROI_mask is not None:
        cv2.bitwise_and(segmented_frame, ROI_mask, segmented_frame)

    # Extract blobs info
    approximation_method = (  # disable contour approximation for small blobs
        cv2.CHAIN_APPROX_SIMPLE if area_ths[0] < 35 else cv2.CHAIN_APPROX_TC89_KCOS
    )
    contours = cv2.findContours(
        segmented_frame, cv2.RETR_EXTERNAL, approximation_method
    )[0]

    # Filter contours by size
    areas = []
    good_contours = []
    for contour in contours:
        area = cv2.contourArea(contour)
        if area_ths[0] <= area <= area_ths[1]:
            good_contours.append(np.squeeze(contour))
            areas.append(area)
    return areas, good_contours, frame


def segment(
    segmentation_parameters: dict,
    episodes: list[Episode],
    bbox_images_dir: Path | None,
    number_of_frames: int,
    n_jobs: int,
) -> list[list[Blob]]:
    """Computes a list of blobs for each frame of the video and the maximum
    number of blobs found in a frame."""
    logging.info(
        f"Segmenting video, {len(episodes)} episodes in {n_jobs} parallel jobs"
    )
    logging.info(f"Saving bounding box images in {bbox_images_dir or 'RAM'}")

    if bbox_images_dir is not None:
        # We indicate if we want bbox on disk or ram by populating episode's bbox_images attribute
        for episode in episodes:
            episode.bbox_images = bbox_images_dir / f"bbox_images_{episode.index}.h5"

    inputs = [(episode, segmentation_parameters) for episode in episodes]

    blobs_in_video: list[list[Blob]] = [[]] * number_of_frames

    if n_jobs == 1:
        for input in track(inputs, "Segmenting video"):
            blobs_in_episode, episode_, bbox_images = segment_episode(input)
            blobs_in_video[episode_.global_start : episode_.global_end] = (
                blobs_in_episode
            )
            # populate episode bbox_images with the process file (BytesIO or disk path)
            episodes[episode_.index].bbox_images = bbox_images
    else:
        with Pool(
            n_jobs,
            maxtasksperchild=3,
            initializer=setup_logging_queue,
            initargs=(LOGGING_QUEUE,),
        ) as p:
            for blobs_in_episode, episode_, bbox_images in track(
                p.imap_unordered(segment_episode, inputs),
                "Segmenting video",
                len(inputs),
            ):
                blobs_in_video[episode_.global_start : episode_.global_end] = (
                    blobs_in_episode
                )

                # populate episode bbox_images with the process file (BytesIO or disk path)
                episodes[episode_.index].bbox_images = bbox_images
    return blobs_in_video


def generate_frame_stack(
    episodes: list[Episode],
    n_frames_for_background: int,
    progress_bar=None,
    abort: Callable = lambda: False,
) -> np.ndarray | None:
    logging.info(
        "Generating frame stack for background subtraction with"
        f" {n_frames_for_background} samples"
    )

    list_of_frames: list[tuple[int, Path]] = []
    for episode in episodes:
        list_of_frames += [
            (frame, episode.video_path)
            for frame in range(episode.local_start, episode.local_end)
        ]

    frames_to_take = np.linspace(
        0, len(list_of_frames) - 1, n_frames_for_background, dtype=int
    )

    frames_to_sample: list[tuple[int, Path]] = [
        list_of_frames[i] for i in frames_to_take
    ]

    cap = cv2.VideoCapture(str(episodes[0].video_path))
    height = int(cap.get(cv2.CAP_PROP_FRAME_HEIGHT))
    width = int(cap.get(cv2.CAP_PROP_FRAME_WIDTH))
    if abort():
        return None
    frame_stack = np.empty((len(frames_to_sample), height, width), np.uint8)
    current_video = 0
    error_frames: list[int] = []
    for i, (frame_number, video_path) in enumerate(
        track(frames_to_sample, "Computing background")
    ):
        if video_path != current_video:
            cap.release()
            cap = cv2.VideoCapture(str(video_path))
            current_video = video_path
        if frame_number != int(cap.get(cv2.CAP_PROP_POS_FRAMES)):
            cap.set(cv2.CAP_PROP_POS_FRAMES, frame_number)
        ret, frame = cap.read()
        if ret:
            frame_stack[i] = to_gray_scale(frame)
        else:
            logging.error(
                f"OpenCV could not read frame {frame_number} of"
                f" {video_path} while computing the background"
            )
            error_frames.append(i)
        if abort():
            return None
        if progress_bar:
            progress_bar.emit(i)

    if error_frames:
        return np.delete(frame_stack, error_frames, 0)
    return frame_stack


def generate_background_from_frame_stack(
    frame_stack: np.ndarray | None, stat: str
) -> np.ndarray | None:
    if frame_stack is None:
        return None

    logging.info(
        f"Computing background from a stack of {len(frame_stack)} frames using '{stat}'"
    )

    if stat == "median":
        bkg = np.median(frame_stack, axis=0, overwrite_input=True)
    elif stat == "mean":
        bkg = frame_stack.mean(0)
    elif stat == "max":
        bkg = frame_stack.max(0)
    elif stat == "min":
        bkg = frame_stack.min(0)
    else:
        raise ValueError(
            f"Stat '{stat}' is not one of ('median', 'mean', 'max' or 'min')"
        )
    return bkg.astype(np.uint8)


def compute_background(
    episodes: list[Episode], n_frames_for_background: int, stat: str, progress_bar=None
) -> np.ndarray | None:
    """
    Computes the background model by sampling `n_frames_for_background` frames
    from the video and computing the stat ('median', 'mean', 'max' or 'min')
    across the sampled frames.

    Parameters
    ----------
    video_paths : list[str]
    original_ROI: np.ndarray
    episodes: list[tuple(int, int, int, int, int)]
    stat: str
        statistic to compute over the sampled frames
        ('median', 'mean', 'max' or 'min')

    Returns
    -------
    bkg : np.ndarray
        Background model
    """

    frame_stack = generate_frame_stack(episodes, n_frames_for_background, progress_bar)

    if frame_stack is None:
        return None

    background = generate_background_from_frame_stack(frame_stack, stat)

    return background


def load_custom_background(
    path: str, example_video_path: Path | str | None = None
) -> np.ndarray:
    logging.info(f"Loading custom background from {path}")
    try:
        bkg = cv2.imdecode(np.fromfile(path, dtype=np.uint8), cv2.IMREAD_COLOR)
        assert bkg is not None
        bkg = cv2.cvtColor(bkg, cv2.COLOR_BGR2GRAY)
    except Exception:
        raise IdtrackeraiError(
            f"Could not read the image file: {path}. Please select a valid"
            " image file."
        )

    if example_video_path is None:
        return bkg

    cap = cv2.VideoCapture(str(example_video_path))
    required_shape = (
        int(cap.get(cv2.CAP_PROP_FRAME_HEIGHT)),
        int(cap.get(cv2.CAP_PROP_FRAME_WIDTH)),
    )
    cap.release()

    if bkg.shape != required_shape:
        raise IdtrackeraiError(
            f"The uploaded background image has shape {bkg.shape}, which does not"
            f" match the required shape {required_shape} based on the video"
            f" ({example_video_path}). Please upload a background image with the"
            " correct dimensions."
        )

    return bkg


def to_gray_scale(frame: np.ndarray) -> np.ndarray:
    if frame.ndim > 2:
        frame = cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY)
    return frame


def get_bbox_image(frame: np.ndarray, cnt: np.ndarray) -> np.ndarray:
    """Computes the `bbox_image`from a given frame and contour. It also
    returns the coordinates of the `bbox`, the ravelled `pixels`
    inside of the contour and the diagonal of the `bbox` as
    an `estimated_body_length`

    Parameters
    ----------
    frame : nd.array
        frame from where to extract the `bbox_image`
    cnt : list
        List of the coordinates that defines the contour of the blob in the
        full frame of the video

    Returns
    -------
    bbox_image : nd.array
        Part of the `frame` defined by the coordinates in `bbox`
    """
    # extra padding for future image eroding
    pad = 1
    # Coordinates of an expanded bounding box
    frame_w, frame_h = frame.shape
    x0, y0, w, h = cv2.boundingRect(cnt)
    w -= 1  # cv2 adds an extra pixel on width and height
    h -= 1
    x0 -= pad
    y0 -= pad
    x1 = x0 + w + 2 * pad
    y1 = y0 + h + 2 * pad

    if x0 < 0:
        x0_margin = -x0
        x0 = 0
    else:
        x0_margin = 0

    if y0 < 0:
        y0_margin = -y0
        y0 = 0
    else:
        y0_margin = 0

    if x1 > frame_h:
        x1_margin = frame_h - x1
        x1 = frame_h
    else:
        x1_margin = None

    if y1 > frame_w:
        y1_margin = frame_w - y1
        y1 = frame_w
    else:
        y1_margin = None

    bbox_image = np.zeros((h + 2 * pad, w + 2 * pad), np.uint8)

    # the estimated body length is the diagonal of the original bbox
    # Get bounding box from frame
    bbox_image[y0_margin:y1_margin, x0_margin:x1_margin] = frame[y0:y1, x0:x1]
    return bbox_image



import logging
from collections.abc import Callable, Iterable
from datetime import timedelta
from operator import length_hint
from pathlib import Path
from typing import IO, TypeVar

from rich.progress import (
    BarColumn,
    DownloadColumn,
    FileSizeColumn,
    Progress,
    TaskProgressColumn,
    TimeRemainingColumn,
    TransferSpeedColumn,
    _Reader,
)

InputType = TypeVar("InputType")


def track(
    sequence: Iterable[InputType],
    desc: str = "Working...",
    total: float | None = None,
    verbose: bool = True,
    callback: Callable[[float], None] | None = None,
) -> Iterable[InputType]:
    """A custom interpretation of rich.progress.track"""

    if not verbose:
        yield from sequence
        return

    progress = Progress(
        "         " + desc,
        BarColumn(bar_width=None),
        TaskProgressColumn(show_speed=True),
        TimeRemainingColumn(elapsed_when_finished=True),
        transient=True,
    )

    with progress:
        if callback is None:
            yield from progress.track(sequence, total, description=desc)
        else:  # used for progress bars in GUIs
            callback_total = total or float(length_hint(sequence)) or 1
            for i, iteration in enumerate(
                progress.track(sequence, total, description=desc), 1
            ):
                yield iteration
                callback(i / callback_total)
            callback(1)

    task = progress.tasks[0]

    logging.info(
        "[green]%s[/] (%s iterations). It took %s",
        desc,
        int(task.total) if task.total is not None else "unknown",
        "--:--" if task.elapsed is None else timedelta(seconds=int(task.elapsed)),
        stacklevel=2,
        extra={"markup": True},
    )


class _ReaderWriter(_Reader):
    """An extension of rich.progress._Reader to allow file writing"""

    def writable(self) -> bool:
        return True

    def write(self, s) -> int:
        block = self.handle.write(s)
        self.progress.advance(self.task, advance=block)
        return block


class open_track:
    """
    Context manager to track progress in large object disk read/write processes.
    So far, only used in :class:`ListOfBlobs` and :class:`ListOfFragments`.
    """

    def __init__(
        self,
        file: str | Path,
        mode: str = "r",
        buffering: int = -1,
        encoding: str | None = None,
        verbose: bool = True,
    ) -> None:

        reading = mode in ("r", "rb", "rt")

        self.desc = (
            ("Reading" if reading else "Writing") + " [italic]" + Path(file).name
        )

        if not verbose:
            self.reader = open(file, mode, buffering, encoding)  # noqa: SIM115
            self.progress = None
            return

        self.progress = Progress(
            "         " + self.desc,
            BarColumn(bar_width=None),
            DownloadColumn() if reading else FileSizeColumn(),
            (
                TimeRemainingColumn(elapsed_when_finished=True)
                if reading
                else TransferSpeedColumn()
            ),
            transient=True,
        )

        if reading:
            self.reader = self.progress.open(
                file, mode, buffering, encoding, description=self.desc
            )
        else:
            self.reader = _ReaderWriter(
                # We are reusing rich.progress._Reader, with works with BinaryIO, with an any IO, lets hope it works on all OS
                open(file, mode, buffering, encoding),  # type: ignore # noqa: SIM115
                self.progress,
                self.progress.add_task(self.desc, total=None),
            )

    def __enter__(self) -> IO:
        if self.progress is not None:
            self.progress.start()
        return self.reader.__enter__()

    def __exit__(self, exc_type, exc_value, traceback) -> None:
        if self.progress is not None:
            self.progress.stop()
            task = self.progress.tasks[0]

            logging.info(
                "[green]%s[/] (%s). It took %s",
                self.desc,
                FileSizeColumn().render(task),
                (
                    "--:--"
                    if task.elapsed is None
                    else timedelta(seconds=int(task.elapsed))
                ),
                stacklevel=2,
                extra={"markup": True},
            )
        self.reader.__exit__(exc_type, exc_value, traceback)



"""isort:skip_file"""

import multiprocessing
import os
import sys
from contextlib import suppress
from importlib import metadata

if "forkserver" in multiprocessing.get_all_start_methods():
    multiprocessing.set_start_method("forkserver", force=True)

with suppress(ImportError):
    # PyQt has to be imported before CV2 (importing idtrackerai stuff implies CV2)
    # If not, the QFileDialog.getFileNames() does not load the icons, very weird

    # lets try to go for PyQt6
    if "QT_API" not in os.environ:
        os.environ["QT_API"] = "pyqt6"
    try:
        from qtpy.QtWidgets import QApplication  # noqa F401
    except ImportError:
        # ups! PyQt6 failed. Lets forget everything and try again
        os.environ.pop("QT_API")
        # in Ubuntu I've seen PyQt6 to remain in sys.modules even though it failed importing
        sys.modules.pop("PyQt6", None)
        from qtpy.QtWidgets import QApplication  # noqa F401

from .utils import IdtrackeraiError, conf

# Video has to be the first class to be imported
from .session import Session

from .blob import Blob
from .fragment import Fragment
from .globalfragment import GlobalFragment
from .list_of_blobs import ListOfBlobs
from .list_of_fragments import ListOfFragments
from .list_of_global_fragments import ListOfGlobalFragments

__version__ = metadata.version("idtrackerai")


__all__ = [
    "Blob",
    "ListOfBlobs",
    "ListOfFragments",
    "ListOfGlobalFragments",
    "ListOfGlobalFragments",
    "GlobalFragment",
    "Session",
    "Fragment",
    "IdtrackeraiError",
    "conf",
]



import logging
from collections.abc import Callable
from colorsys import hsv_to_rgb
from itertools import pairwise
from pathlib import Path

import cv2
import numpy as np
from qtpy.QtCore import Qt
from qtpy.QtGui import QColor, QImage, QPainter
from qtpy.sip import voidptr

from idtrackerai import Session
from idtrackerai.GUI_tools import VideoPathHolder
from idtrackerai.utils import load_trajectories, track


def _QImageToArray(qimg: QImage) -> np.ndarray:
    width = qimg.width()
    height = qimg.height()
    byte_str: voidptr = qimg.bits()  # type: ignore
    return np.asarray(byte_str.asarray(height * width * 4)).reshape(height, width, 4)[
        ..., :-1
    ]


def draw_general_frame(
    np_frame: np.ndarray,
    frame_number: int,
    trajectories: np.ndarray,
    centroid_trace_length: int,
    colors: list[tuple[int, ...]] | list[QColor],
    labels: list[str] | None,
) -> np.ndarray:
    ordered_centroid = trajectories[frame_number]
    match np_frame.ndim:
        case 3:
            frame = QImage(
                np_frame.data,
                np_frame.shape[1],
                np_frame.shape[0],
                np_frame.shape[1] * 3,
                QImage.Format.Format_BGR888,
            )
        case 2:
            frame = QImage(
                np_frame.data,
                np_frame.shape[1],
                np_frame.shape[0],
                np_frame.shape[1],
                QImage.Format.Format_Grayscale8,
            )
        case _:
            raise RuntimeError("Invalid frame shape ", np_frame.shape)
    canvas = QImage(frame.size(), QImage.Format.Format_ARGB32_Premultiplied)
    canvas.fill(Qt.GlobalColor.transparent)
    painter = QPainter(canvas)
    painter.setRenderHint(QPainter.RenderHint.Antialiasing, True)
    painter.setCompositionMode(QPainter.CompositionMode.CompositionMode_Source)
    pen = painter.pen()
    pen.setWidth(2)
    for cur_id, centroid in enumerate(ordered_centroid):
        if frame_number > centroid_trace_length:
            centroids_trace = trajectories[
                frame_number - centroid_trace_length : frame_number + 1, cur_id
            ]
        else:
            centroids_trace = trajectories[: frame_number + 1, cur_id]

        color = colors[cur_id]
        if not isinstance(color, QColor):
            color = QColor(*color)

        alphas = np.linspace(0, 255, len(centroids_trace), dtype=int)[1:]
        if len(centroids_trace) > 1:
            for alpha, (pointA, pointB) in zip(alphas, pairwise(centroids_trace)):
                if any(pointA < 0) or any(pointB < 0):
                    continue
                color.setAlpha(alpha)
                pen.setColor(color)
                painter.setPen(pen)
                painter.drawLine(*pointA, *pointB)

        if all(centroid > 0):
            color.setAlpha(255)
            painter.setBrush(color)
            painter.setPen(Qt.PenStyle.NoPen)
            painter.drawEllipse(centroid[0] - 3, centroid[1] - 3, 6, 6)

    painter.setCompositionMode(QPainter.CompositionMode.CompositionMode_DestinationOver)
    painter.drawImage(canvas.rect(), frame)
    painter.end()

    arr_img = _QImageToArray(canvas).copy()
    if not labels:
        return arr_img

    for cur_id, centroid in enumerate(ordered_centroid):
        if all(centroid > 0):
            color = colors[cur_id]
            if isinstance(color, QColor):
                color = (color.blue(), color.green(), color.red())
            else:
                color = (color[2], color[1], color[0])

            arr_img = cv2.putText(
                arr_img,
                labels[cur_id],
                (centroid[0], centroid[1]),
                fontFace=cv2.FONT_HERSHEY_SIMPLEX,
                fontScale=0.8,
                color=color,
                thickness=2,
            )

    return arr_img


def generate_general_video(
    session: Session,
    trajectories_path: Path | str | None = None,
    draw_in_gray: bool = False,
    centroid_trace_length: int = 20,
    starting_frame: int = 0,
    ending_frame: int | None = None,
    labels: list[str] | None = None,
    resize_factor: float | None = None,
    callback: Callable | None = None,
    output_path: Path | str | None = None,
    colors: list[tuple[int, ...]] | None = None,
) -> None:
    """
    Generates a video with animal trajectories overlaid on the original video frames.

    This function creates a new video file with the tracked trajectories of each animal
    drawn on top of the video frames. Optionally, the video can be rendered in grayscale,
    resized, and include labels for each animal. The output video is saved in the session folder
    with the suffix '_tracked.avi'.

    .. seealso::
        :ref:`video generator` in the documentation for more details.

    Parameters
    ----------
    session : Session
        The idtrackerai Session object containing video and experiment information.
    trajectories_path : Path | str | None, optional
        Path to the trajectories file. If None, trajectories are loaded from the session folder.
    draw_in_gray : bool, optional
        If True, the output video will be in grayscale. Default is False (color).
    centroid_trace_length : int, optional
        Number of previous frames to show as a trace for each animal's trajectory. Default is 20.
    starting_frame : int, optional
        Index of the first frame to include in the output video. Default is 0.
    ending_frame : int | None, optional
        Index of the last frame to include in the output video. If None, uses the last frame available.
    labels : list[str] | None, optional
        List of labels to display for each animal. If None, no labels are shown.
    resize_factor : float | None, optional
        Factor by which to resize the video frames. If None, automatically scales to fit within 1920x1080.
    callback : Callable | None, optional
        Optional callback function for progress tracking (e.g., for GUI updates).
    output_path : Path | str | None, optional
        Path where the output video will be saved. If None, saves in the session folder with the name '<original_video_name>_tracked.avi'.

    Notes
    -----
    The output video is saved in the session folder with the name '<original_video_name>_tracked.avi'.
    """
    if draw_in_gray:
        logging.info("Drawing original video in grayscale")

    if resize_factor is None:
        resize_factor = min(1920 / session.width, 1080 / session.height, 1)

    if resize_factor != 1:
        logging.info(f"Applying resize of factor {resize_factor}")

    trajectories = load_trajectories(trajectories_path or session.trajectories_folder)[
        "trajectories"
    ]
    trajectories = np.nan_to_num(trajectories * resize_factor, nan=-100).astype(int)

    if colors is None:
        if session.identities_colors:
            colors = [QColor(c).getRgb()[:3] for c in session.identities_colors]  # type: ignore
            assert colors is not None
        else:
            colors = [
                tuple(int(v * 255) for v in hsv_to_rgb(h / session.n_animals, 1, 1))
                for h in range(session.n_animals)
            ]

    path_to_save_video = output_path or (
        session.session_folder / (session.video_paths[0].stem + "_tracked.avi")
    )

    out_video_width = int(session.width * resize_factor + 0.5)
    out_video_height = int(session.height * resize_factor + 0.5)

    video_writer = cv2.VideoWriter(
        str(path_to_save_video),
        cv2.VideoWriter.fourcc(*"XVID"),
        session.frames_per_second,
        (out_video_width, out_video_height),
    )

    videoPathHolder = VideoPathHolder(session.video_paths)

    ending_frame = len(trajectories) - 1 if ending_frame is None else ending_frame
    logging.info(f"Drawing from frame {starting_frame} to {ending_frame}")

    for frame in track(
        range(starting_frame, ending_frame), "Generating video", callback=callback
    ):
        try:
            img = videoPathHolder.read_frame(frame, not draw_in_gray)
            if resize_factor != 1:
                img = cv2.resize(img, (0, 0), fx=resize_factor, fy=resize_factor)
        except RuntimeError as exc:
            logging.error(str(exc))
            img = np.zeros(
                (
                    (out_video_height, out_video_width)
                    if draw_in_gray
                    else (out_video_height, out_video_width, 3)
                ),
                np.uint8,
            )

        img = draw_general_frame(
            img, frame, trajectories, centroid_trace_length, colors, labels
        )

        video_writer.write(img)

    logging.info(f"Video generated in {path_to_save_video}")



import json
import logging
from pathlib import Path

from idtrackerai import (
    GlobalFragment,
    ListOfFragments,
    ListOfGlobalFragments,
    Session,
    __version__,
    conf,
)
from idtrackerai.utils import create_dir

from ..network import DEVICE, IdCNN, IdentifierBase, IdentifierIdCNN
from .accumulation_manager import AccumulationManager
from .accumulator import accumulation_step
from .assigner import assign_remaining_fragments, check_penultimate_model
from .contrastive import ContrastiveLearning, IdentifierContrastive
from .identity_transfer import identify_first_global_fragment_for_accumulation


def run_tracker(
    session: Session,
    list_of_fragments: ListOfFragments,
    list_of_global_fragments: ListOfGlobalFragments,
) -> IdentifierBase:
    logging.info("Tracking with identities")
    create_dir(session.accumulation_folder, remove_existing=True)

    with session.new_timer("Fragment identification"):
        identifier_model = fragment_identification(
            session, list_of_fragments, list_of_global_fragments
        )
        identifier_model.save(session.accumulation_folder)

    with (session.accumulation_folder / "model_params.json").open("w") as file:
        json.dump(
            {
                "n_classes": session.n_animals,
                "image_size": session.id_image_size,
                "resolution_reduction": session.resolution_reduction,
                "model": identifier_model.model.__class__.__name__,
                "version": __version__,
            },
            file,
            indent=4,
        )

    with session.new_timer("Identification"):
        assign_remaining_fragments(list_of_fragments, identifier_model)

    return identifier_model


def fragment_identification(
    session: Session,
    list_of_fragments: ListOfFragments,
    list_of_global_fragments: ListOfGlobalFragments,
) -> IdentifierBase:

    list_of_fragments.reset(roll_back_to="fragmentation")

    # Instantiate accumulation manager
    accumulation_manager = AccumulationManager(
        session.n_animals, list_of_fragments, list_of_global_fragments
    )

    first_global_fragment = (
        max(list_of_global_fragments, key=lambda gf: gf.minimum_distance_travelled)
        if list_of_global_fragments.global_fragments
        else None
    )

    if first_global_fragment is None:
        logging.info("The video does not contain any long enough Global Fragment")
        if session.exclusive_rois:
            logging.warning(  # TODO
                "Right now it is not possible to have exclusive ROIs without Global Fragments. We are working on it"
            )
    else:
        session.first_frame_first_global_fragment = (
            first_global_fragment.first_frame_of_the_core
        )
        identify_first_global_fragment_for_accumulation(
            first_global_fragment,
            session,
            session.knowledge_transfer_folder,
            session.id_image_size,
        )
        session.identities_groups = list_of_fragments.build_exclusive_rois()

    list_of_global_fragments.sort_by_distance_to_the_frame(
        session.first_frame_first_global_fragment
    )

    if conf.DISABLE_CONTRASTIVE:
        logging.warning("Contrastive step is disabled")
    else:
        with session.new_timer("Contrastive step"):
            identifier_contrastive, ratio_accumulated = contrastive_step(
                first_global_fragment,
                session.knowledge_transfer_folder,
                list_of_fragments,
                session,
                accumulation_manager,
            )
            session.ratio_accumulated_images = ratio_accumulated

            if ratio_accumulated == float("inf"):
                logging.info(
                    "There are no Global Fragments for an accumulation protocol\n"
                    "[bold]We will not run the accumulation protocol[/].",
                    extra={"markup": True},
                )
                return identifier_contrastive

            logging.info(
                f"Contrastive step identified {ratio_accumulated:.2%} of the accumulable images"
            )
            CONTRASTIVE_MIN_ACCUMULATION = conf.CONTRASTIVE_MIN_ACCUMULATION
            if ratio_accumulated >= CONTRASTIVE_MIN_ACCUMULATION:
                logging.info(
                    f"This is higher than {CONTRASTIVE_MIN_ACCUMULATION=:.1%}, "
                    "enough to finish accumulation right here.\n"
                    "[bold]We will not run the accumulation protocol[/].",
                    extra={"markup": True},
                )

                # if we do not run the accumulation protocol, we need to set the final
                # identities to the fragments accumulated by contrastive
                for frag in list_of_fragments:
                    if frag.acceptable_for_training and not frag.used_for_training:
                        assert frag.temporary_id is not None
                        frag.used_for_training = True
                        frag.acceptable_for_training = False
                        frag.accumulated_globally = True
                        frag.accumulation_step = 0
                        frag.identity = frag.temporary_id + 1
                        frag.P1_vector[:] = 0.0
                        frag.P1_vector[frag.temporary_id] = 1.0

                return identifier_contrastive

            logging.info(
                f"This is lower than {CONTRASTIVE_MIN_ACCUMULATION=:.1%}, "
                "[bold]not[/] enough to finish accumulation right here.\n"
                "[bold]We will run the accumulation protocol[/].",
                extra={"markup": True},
            )

            if ratio_accumulated < 0.8 and session.silhouette_score is not None:
                if session.silhouette_score > conf.CONTRASTIVE_SILHOUETTE_TARGET:
                    logging.warning(
                        "Such a low ratio of accumulated images with a Silhouette score "
                        "above the target may indicate the need to increase such target "
                        "or to check again the segmentation parameters"
                    )
                else:
                    logging.warning(
                        "Such a low ratio of accumulated images with a Silhouette score "
                        "below the target may indicate the need to increase the training "
                        "patience or to check again the segmentation parameters"
                    )

    if session.knowledge_transfer_folder:
        try:
            identification_cnn = IdCNN.load(
                session.id_image_size, session.knowledge_transfer_folder
            ).to(DEVICE)
        except FileNotFoundError:
            logging.warning(
                "IdCNN model not found in the knowledge transfer folder "
                f'"{session.knowledge_transfer_folder}", proceeding with a randomly initialized model'
            )
            identification_cnn = IdCNN(session.id_image_size, session.n_animals).to(
                DEVICE
            )
    else:
        identification_cnn = IdCNN(session.id_image_size, session.n_animals).to(DEVICE)

    if conf.TORCH_COMPILE:
        identification_cnn.compile()

    model_path = session.accumulation_folder / "tmp_identification_network.pt"
    penultimate_model_path = session.accumulation_folder / (
        "tmp_identification_network_penultimate.pt"
    )

    with session.new_timer("Accumulation protocol"):
        while accumulation_manager.new_global_fragments_for_training:
            early_stopped = accumulation_step(
                accumulation_manager,
                session,
                identification_cnn,
                model_path,
                penultimate_model_path,
            )
            if early_stopped:
                logging.info("We don't need to accumulate more images")
                break
        else:
            logging.info("No more new images to accumulate")

    ratio_accumulated = accumulation_manager.ratio_accumulated_images
    session.ratio_accumulated_images = ratio_accumulated

    if ratio_accumulated > 0.9:
        logging.info(
            f"[green]We accumulated enough images ({ratio_accumulated:.2%}). Accumulation protocol succeeded!",
            extra={"markup": True},
        )
    else:
        logging.warning(
            f"[red]We did not accumulate enough images ({ratio_accumulated:.2%}). Accumulation protocol failed!",
            extra={"markup": True},
        )

    check_penultimate_model(identification_cnn, model_path, penultimate_model_path)
    model_path.unlink(missing_ok=True)
    model_path.with_suffix(".metadata.json").unlink(missing_ok=True)
    penultimate_model_path.unlink(missing_ok=True)
    penultimate_model_path.with_suffix(".metadata.json").unlink(missing_ok=True)

    return IdentifierIdCNN(identification_cnn)


def contrastive_step(
    first_global_fragment: GlobalFragment | None,
    knowledge_transfer_folder: Path | None,
    list_of_fragments: ListOfFragments,
    session: Session,
    accumulation_manager: AccumulationManager,
) -> tuple[IdentifierContrastive, float]:
    connectivity = list_of_fragments.get_connectivity()
    if connectivity < 0.5:
        logging.warning(
            f"Low fragment connectivity detected: {connectivity:.2f}. The animals in "
            "the video appear too isolated. idtracker.ai relies on observing groups of "
            "animals visible at the same time to effectively train the model. Limited "
            "coexistence may reduce tracking accuracy."
        )
    else:
        logging.info(
            f"Fragment connectivity is {connectivity:.2f}, which is good enough for "
            "contrastive training"
        )
    session.fragment_connectivity = connectivity

    contrastive = ContrastiveLearning(
        list_of_fragments,
        saving_folder=session.accumulation_folder,
        first_gfrag=first_global_fragment,
    )
    try:
        contrastive.set_model(knowledge_transfer_folder)
    except FileNotFoundError as exc:
        logging.error(exc)
        contrastive.set_model()

    session.silhouette_score = contrastive.train()
    contrastive.predict(list_of_fragments, first_global_fragment)

    if not list_of_fragments.n_images_in_global_fragments:
        # there are no global fragments
        contrastive.model_checkpoint_path.unlink()
        return contrastive.get_identification_model(), float("inf")

    accumulation_manager.assign_identities()
    accumulation_manager.update_accumulation_statistics()
    session.accumulation_statistics_data = accumulation_manager.accumulation_statistics

    n_accumulated_images = sum(
        fragment.n_images
        for fragment in list_of_fragments.individual_fragments
        if fragment.acceptable_for_training and not fragment.used_for_training
    )

    ratio = n_accumulated_images / list_of_fragments.n_images_in_global_fragments

    if ratio >= conf.CONTRASTIVE_MIN_ACCUMULATION:
        # remove contrastive checkpoint because the whole IdentifierContrastive will be saved instead
        contrastive.model_checkpoint_path.unlink(missing_ok=True)

    return contrastive.get_identification_model(), ratio



"""Module to validate the tracking data of a session, see :ref:`validator_reference`."""

import sys
from argparse import ArgumentParser
from multiprocessing import Process
from pathlib import Path

from qtpy.QtWidgets import QApplication

from idtrackerai import Session
from idtrackerai.utils import (
    LOGGING_QUEUE,
    manage_exception,
    setup_logging_queue,
    wrap_entrypoint,
)

from .validation_GUI import ValidationGUI


@wrap_entrypoint
def idtrackerai_validate_entrypoint() -> None:
    parser = ArgumentParser()
    parser.add_argument(
        "session_directory", help="Session directory to validate", type=Path, nargs="?"
    )
    args = parser.parse_args()

    idtrackerai_validate(args.session_directory)


def idtrackerai_validate(session_directory: Session | Path | str | None = None) -> None:
    """Initialize the Validator GUI, called by the command ``idtrackerai_validate``.

    .. seealso::
        Documentation for :ref:`validator_reference`

    Parameters
    ----------
    session_directory : Session | Path | str | None, optional
        Session or path to the Session folder to validate. If None, a blank Validator is opened.
    """
    if sys.platform.startswith("win") or sys.platform == "darwin":
        run_gui(session_directory, None)
    else:
        # On Linux, we run the GUI in a separate process to catch libxcb errors
        p = Process(target=run_gui, args=(session_directory, LOGGING_QUEUE))
        p.start()
        p.join()

        if p.exitcode != 0:
            raise RuntimeError(f"QApplication crashed with exit code {p.exitcode}")


def run_gui(session_directory: Session | Path | str | None, logging_queue):
    if logging_queue:
        setup_logging_queue(logging_queue)

    # this catches exceptions when raised inside Qt
    def excepthook(exc_type, exc_value, exc_tb) -> None:
        QApplication.quit()
        manage_exception(exc_value)

    sys.excepthook = excepthook
    app = QApplication(sys.argv)
    window = ValidationGUI()
    window.show()
    if session_directory is not None:
        window.open_session(session_directory)
    else:
        window.ask_for_session()
    app.exec()



"""Module to report usage analytics and check for updates."""

import json
import logging
import os
import re
import sys
from contextlib import suppress
from datetime import datetime
from enum import Enum
from functools import lru_cache
from importlib import metadata
from pathlib import Path
from platform import platform, python_version
from threading import Thread
from urllib.request import urlopen

from requests import post

ANALYTICS_STATE_FILE_PATH = Path(__file__).parent / "usage_analytics_state.json"
ANALYTICS_URL = "https://analytics.polaviejalab.org/report_usage.php"
PYPI_URL = "https://pypi.org/simple/idtrackerai"
ANALYTICS_ENVIRON = "IDTRACKERAI_DISABLE_ANALYTICS"


class ComparisonResult(Enum):
    EQUAL = "equal"
    MAJOR_UPDATE = "major_update"
    MINOR_UPDATE = "minor_update"
    PATCH_UPDATE = "patch_update"
    STABLE_RELEASE = "stable_release"
    ERROR = "error"

    @property
    def update_available(self) -> bool:
        return self in {
            ComparisonResult.MAJOR_UPDATE,
            ComparisonResult.MINOR_UPDATE,
            ComparisonResult.PATCH_UPDATE,
        }


def check_version_on_console_thread() -> None:
    Thread(target=check_version_on_console).start()


def report_usage_on_console_thread() -> None:
    Thread(target=report_usage).start()


def set_usage_analytics_state(enabled: bool) -> None:
    ANALYTICS_STATE_FILE_PATH.write_text(json.dumps(enabled))


def get_usage_analytics_state() -> bool:
    """Returns the current state of the usage
    analytics reporting. If the state is not set,
    it will return True and set the state to True."""
    environ = os.environ.get(ANALYTICS_ENVIRON, "").lower()
    if environ in ("1", "true"):
        state = False
    elif environ in ("0", "false") or not ANALYTICS_STATE_FILE_PATH.exists():
        state = True
    else:
        state = json.loads(ANALYTICS_STATE_FILE_PATH.read_text())

    current_state = (
        json.loads(ANALYTICS_STATE_FILE_PATH.read_text())
        if ANALYTICS_STATE_FILE_PATH.exists()
        else None
    )
    if state != current_state:
        set_usage_analytics_state(state)

    return state


def report_usage() -> None:
    """Reports usage analytics to the server."""
    usage_analytics_enabled = get_usage_analytics_state()
    if not usage_analytics_enabled:
        logging.info("Usage analytics reporting is disabled")
        return

    try:
        response = post(
            ANALYTICS_URL,
            json={
                "date": datetime.now().astimezone().isoformat(),
                "platform": platform(True),
                "idtrackerai_version": metadata.version("idtrackerai"),
                "python_version": python_version(),
                "command": sys.argv,
            },
        )
        if response.status_code != 200:
            logging.error(
                f"Error reporting usage analytics. Status code: {response.status_code} {response.text}"
            )
    except Exception as e:
        logging.error(f"Error reporting usage analytics: {e}")


def check_version_on_console() -> None:
    logger = logging.getLogger()
    old_level = logger.getEffectiveLevel()
    logger.setLevel(logging.INFO)
    with suppress(Exception):
        kind, message = check_version()
        if kind == ComparisonResult.ERROR:
            logging.error(message)
        elif kind != ComparisonResult.EQUAL:
            logging.warning(message)
    logger.setLevel(old_level)


@lru_cache(maxsize=1)
def check_version() -> tuple[ComparisonResult, str]:
    """Check if there is a new version of idtracker.ai available."""
    try:
        out_text = urlopen(PYPI_URL, timeout=5).read().decode("utf-8")
    except Exception:
        return ComparisonResult.ERROR, "Error fetching PyPI data"

    if not isinstance(out_text, str) or not out_text:
        return ComparisonResult.ERROR, "Error reading PyPI data"

    no_yanked_versions = "\n".join(
        line for line in out_text.splitlines() if "yanked" not in line
    )
    matches: list[tuple[str, str]] = re.findall(
        ">idtrackerai-(.+?)(.tar.gz|-py3-none-any.whl)<", no_yanked_versions
    )

    current_version_str = metadata.version("idtrackerai")
    current_version = current_version_str.split("a")[0]
    try:
        current_version = tuple(map(int, current_version.split(".")))
    except Exception as e:
        logging.error(f"Error parsing current version: {e}")
        return ComparisonResult.ERROR, (
            f"The current version of idtracker.ai ({current_version}) is not a "
            "valid version string."
        )

    latest_version = max(
        tuple(map(int, version.split(".")))
        for version, _file_extension in matches
        if version.replace(".", "").isdigit()  # only keep stable versions
    )

    latest_version_str = ".".join(map(str, latest_version))

    try:
        if latest_version[0] > current_version[0]:
            return ComparisonResult.MAJOR_UPDATE, (
                f"A new major release of idtracker.ai is available: {current_version_str} -> "
                f"{latest_version_str}\n"
                'To update, run: "python -m pip install --upgrade idtrackerai"'
            )
        elif latest_version[1] > current_version[1]:
            return ComparisonResult.MINOR_UPDATE, (
                f"A new minor release of idtracker.ai is available: {current_version_str} -> "
                f"{latest_version_str}\n"
                'To update, run: "python -m pip install --upgrade idtrackerai"'
            )
        elif latest_version[2] > current_version[2]:
            return ComparisonResult.PATCH_UPDATE, (
                f"A new patch release of idtracker.ai is available: {current_version_str} -> "
                f"{latest_version_str}\n"
                'To update, run: "python -m pip install --upgrade idtrackerai"'
            )
        elif "a" in current_version_str:
            return ComparisonResult.STABLE_RELEASE, (
                "You are running an alpha version of idtracker.ai and the stable"
                f" version is available: {metadata.version('idtrackerai')} ->"
                f" {latest_version_str}\nTo update, run: python -m pip install --upgrade"
                " idtrackerai"
            )
        else:
            return ComparisonResult.EQUAL, (
                "There are currently no updates available.\n"
                f"Current idtrackerai version: {current_version_str}"
            )
    except IndexError as exc:
        return ComparisonResult.ERROR, (
            f"Error comparing versions {latest_version_str} and {current_version}: {exc}"
        )



import logging
import sys
from collections.abc import Sequence
from datetime import datetime
from enum import Enum
from pathlib import Path
from time import sleep

import numpy as np
import toml
from qtpy.QtCore import Signal  # type: ignore[reportPrivateImportUsage]
from qtpy.QtCore import Qt, QThread
from qtpy.QtGui import QAction, QCloseEvent, QColor, QIcon, QKeyEvent
from qtpy.QtWidgets import (
    QCheckBox,
    QDialog,
    QFileDialog,
    QHBoxLayout,
    QInputDialog,
    QLabel,
    QMessageBox,
    QProgressDialog,
    QPushButton,
    QSpinBox,
    QSplitter,
    QTabWidget,
    QVBoxLayout,
    QWidget,
)

from idtrackerai import Blob, Fragment, ListOfBlobs, ListOfFragments, Session
from idtrackerai.base.postprocess import produce_output_dict
from idtrackerai.GUI_tools import (
    CanvasMouseEvent,
    CanvasPainter,
    GUIBase,
    IdLabels,
    LabelRangeSlider,
    LightPopUp,
    QHLine,
    TransparentDisabledOverlay,
    VideoPlayer,
    build_ROI_patches_from_list,
    get_icon,
    open_session,
)
from idtrackerai.utils import save_trajectories, track

from .widgets import (
    AdditionalInfo,
    ErrorsExplorer,
    IdGroups,
    Interpolator,
    LengthCalibrator,
    MarkMetadata,
    SetupPoints,
    find_selected_blob,
    paintBlobs,
    paintTrails,
)

SELECT_POINT_DIST = 300


class WarningRedirector(logging.Handler):
    def __init__(self, parent: GUIBase) -> None:
        super().__init__()
        self.parent = parent
        self.setLevel(logging.WARNING)

    def emit(self, record: logging.LogRecord) -> None:
        QMessageBox.warning(
            self.parent, "idtracker.ai validator warning", record.getMessage()
        )


class DblClickDialog(QDialog):
    class Answers(Enum):
        Cancel = 0
        ChangeId = 1
        Interpolate = 2
        Reset = 3
        Remove = 4

    def __init__(self, parent: QWidget, n_animals: int) -> None:
        super().__init__(parent)
        self.spinbox = QSpinBox()
        self.spinbox.setMaximum(n_animals)
        main_layout = QVBoxLayout()
        self.setLayout(main_layout)
        self.spinbox.setSpecialValueText("Null identity")

        self.propagate = QCheckBox("Propagate action")
        self.propagate.setChecked(True)
        spin_row = QHBoxLayout()
        spin_row.setAlignment(Qt.AlignmentFlag.AlignCenter)
        spin_row.addWidget(QLabel("New identity:"))
        spin_row.addWidget(self.spinbox)

        cancel_btn = QPushButton(get_icon("cancel"), "Cancel")
        change_id_btn = QPushButton(get_icon("ok"), "Change id")
        remove_btn = QPushButton(QIcon.fromTheme("user-trash"), "Remove\ncentroid")
        reset_id_btn = QPushButton(get_icon("refresh"), "Reset id")
        self.interp_btn = QPushButton("Start interpolation")
        self.interp_btn.setIcon(get_icon("run"))
        first_btn_row = QHBoxLayout()
        first_btn_row.addWidget(remove_btn)
        first_btn_row.addWidget(reset_id_btn)
        first_btn_row.addWidget(change_id_btn)
        second_btn_row = QHBoxLayout()
        second_btn_row.addWidget(self.interp_btn)
        second_btn_row.addWidget(cancel_btn)
        change_id_btn.setDefault(True)

        cancel_btn.clicked.connect(lambda: self.done(self.Answers.Cancel.value))
        change_id_btn.clicked.connect(lambda: self.done(self.Answers.ChangeId.value))
        reset_id_btn.clicked.connect(lambda: self.done(self.Answers.Reset.value))
        remove_btn.clicked.connect(lambda: self.done(self.Answers.Remove.value))
        self.interp_btn.clicked.connect(
            lambda: self.done(self.Answers.Interpolate.value)
        )

        main_layout.addLayout(spin_row)
        main_layout.addWidget(self.propagate, alignment=Qt.AlignmentFlag.AlignCenter)
        main_layout.addLayout(first_btn_row)
        main_layout.addWidget(QHLine())
        main_layout.addLayout(second_btn_row)

    def exec_with_description(
        self, default: int | None
    ) -> tuple[Answers, int | None, bool]:
        if default is not None:
            self.spinbox.setValue(default)
        self.interp_btn.setEnabled(bool(default is not None and default > 0))

        self.spinbox.setFocus()
        answer = self.Answers(super().exec())

        return answer, self.spinbox.value(), self.propagate.isChecked()


class LoadSessionObjects(QThread):
    """Independent thread to load lists of Blobs/Fragments
    because they take too long for large sessions."""

    blobs: ListOfBlobs | None = None
    fragments: list[Fragment] | None = None
    loaded_from: Path
    "Original file of loaded ListOfBlobs. Validated list will be saved in the same location"

    def __init__(self, session: Session, parent: QWidget) -> None:
        super().__init__(parent)
        self.session = session

    def run(self) -> None:
        # when loading light session from CLI, the main windows remains out of focus.
        # This sleeps fixes it, not beautiful but it works...
        sleep(0.1)
        for path in (
            self.session.preprocessing_folder / "list_of_blobs_validated.pickle",
            self.session.preprocessing_folder / "list_of_blobs_no_gaps.pickle",
            self.session.blobs_path,
        ):
            try:
                self.blobs = ListOfBlobs.load(path)
                self.loaded_from = path
                break
            except FileNotFoundError:
                pass
        else:
            logging.warning(
                "List of blobs not found in %s", self.session.blobs_path.parent
            )
            self.blobs = None
        try:
            self.fragments = ListOfFragments.load(
                self.session.fragments_path, reconnect=False
            ).fragments
            for index, fragment in enumerate(self.fragments):
                if fragment.identifier != index:
                    logging.warning(
                        "Loading an old session, invalid list of fragments format"
                    )
                    raise FileExistsError
        except FileNotFoundError:
            self.fragments = None


class SaveSessionObjects(QThread):
    """Independent thread to save lists of Blobs/Fragments
    because they take too long for large sessions."""

    loaded_from: Path
    "Original file of loaded ListOfBlobs. Validated list will be saved in the same location"

    def __init__(
        self, session: Session, blobs: ListOfBlobs, parent: QWidget, loaded_from: Path
    ) -> None:
        super().__init__(parent)
        self.session = session
        self.loaded_from = loaded_from
        self.blobs = blobs

    def run(self) -> None:
        # when loading light session from CLI, the main windows remains out of focus.
        # This sleeps fixes it, not beautiful but it works...
        sleep(0.1)
        self.session.save()
        self.blobs.save(self.loaded_from)


class ValidationGUI(GUIBase):
    blobs: ListOfBlobs

    def __init__(self) -> None:
        super().__init__()

        # TODO logging.getLogger().addHandler(WarningRedirector(self))
        self.light_popup = LightPopUp()
        self.setWindowTitle("Validator")
        self.documentation_url = "https://idtracker.ai/latest/user_guide/validator.html"

        self.video_player = VideoPlayer(self)
        self.widgets_to_close.append(self.video_player)

        self.video_player.canvas.click_event.connect(self.click_on_canvas)
        self.video_player.canvas.double_click_event.connect(self.double_click_on_canvas)

        def new_changes() -> None:
            self.unsaved_changes = True

        self.id_groups = IdGroups(self)
        self.id_groups.needToDraw.connect(self.video_player.update)
        self.id_groups.unsaved_changes.connect(new_changes)

        self.errorsExplorer = ErrorsExplorer()
        self.errorsExplorer.go_to_error.connect(self.go_to_error)

        self.mark_metadata = MarkMetadata(self)
        self.mark_metadata.needToDraw.connect(self.video_player.update)

        self.interpolator = Interpolator()
        self.interpolator.neew_to_draw.connect(self.video_player.update)
        self.interpolator.update_trajectories.connect(self.update_trajectories_range)
        self.interpolator.go_to_frame.connect(self.video_player.setCurrentFrame)
        self.interpolator.preload_frames.connect(self.video_player.preload_frames)
        self.interpolator.interpolation_accepted.connect(
            self.errorsExplorer.accepted_interpolation
        )

        self.id_labels = IdLabels()
        self.id_labels.needToDraw.connect(self.video_player.update)
        self.id_labels.needToDraw.connect(new_changes)

        self.setup_points = SetupPoints()
        self.setup_points.needToDraw.connect(self.video_player.update)
        self.setup_points.needToDraw.connect(new_changes)

        self.length_calibrator = LengthCalibrator()
        self.length_calibrator.needToDraw.connect(self.video_player.update)
        self.length_calibrator.needToDraw.connect(new_changes)

        self.video_player.canvas.click_event.connect(self.setup_points.click_event)
        self.video_player.canvas.click_event.connect(self.length_calibrator.click_event)
        self.video_player.canvas.click_event.connect(self.interpolator.click_event)

        right_splitter = QSplitter(Qt.Orientation.Vertical)
        right_splitter.setContentsMargins(8, 0, 0, 0)

        self.additional_info = AdditionalInfo(self)

        tabs = QTabWidget()
        tabs.setMovable(True)
        tabs.addTab(self.id_groups, "Groups")
        tabs.addTab(self.id_labels, "Labels")
        tabs.addTab(self.setup_points, "Setup Points")
        tabs.addTab(self.length_calibrator, "Length Calibration")
        tabs.addTab(self.mark_metadata, "Mark Metadata")
        TransparentDisabledOverlay(
            "Disable the Interpolator to\nenable these extra tools", tabs
        )

        self.additional_info.metadata_visibility.toggled.connect(
            lambda checked: tabs.setTabVisible(
                tabs.indexOf(self.mark_metadata), checked
            )
        )
        tabs.setTabVisible(
            tabs.indexOf(self.mark_metadata),
            self.additional_info.metadata_visibility.isChecked(),
        )

        right_splitter.setMinimumWidth(250)
        tabs.currentChanged.connect(self.video_player.update)
        right_splitter.addWidget(tabs)
        right_splitter.setHandleWidth(16)
        right_splitter.addWidget(self.additional_info)
        self.interpolator.enabled_changed.connect(
            lambda enabled: tabs.setEnabled(not enabled)
        )

        self.respect_tracking_intervals = QCheckBox("Respect tracking intervals")
        self.respect_tracking_intervals.setChecked(True)
        self.respect_tracking_intervals.stateChanged.connect(
            self.respect_tracking_intervals_changed
        )

        left_splitter = QVBoxLayout()
        left_splitter.addWidget(self.errorsExplorer)
        left_splitter.addWidget(self.respect_tracking_intervals)
        left_splitter.addWidget(QHLine())
        left_splitter.addWidget(self.interpolator)

        splitter = QSplitter(Qt.Orientation.Horizontal, self)
        self.video_player.layout().setContentsMargins(8, 0, 8, 0)
        left_widget = QWidget()
        left_widget.setLayout(left_splitter)
        splitter.addWidget(left_widget)
        splitter.addWidget(self.video_player)
        splitter.addWidget(right_splitter)
        splitter.setStretchFactor(0, 0)
        splitter.setStretchFactor(1, 1)
        splitter.setStretchFactor(2, 0)
        splitter.setSizes((1, 10, 1))
        self.centralWidget().layout().addWidget(splitter)
        self.centralWidget().setEnabled(False)
        self.centralWidget().layout().setContentsMargins(8, 0, 8, 8)

        self.selected_id: int | None = None
        self.selected_blob: Blob | None = None
        self.selection_last_location: tuple[float, float] | None = None

        self.video_player.painting_time.connect(self.paint)
        self.current_frame_number = -1
        self.trajectories: np.ndarray
        """Float, positions of animals"""
        self.unidentified: np.ndarray
        """Bool, there is some identity without centroid"""
        self.duplicated: np.ndarray
        """Bool, some centroid have the same identity"""

        session_menu = self.menuBar().addMenu("Session")
        assert session_menu is not None

        open_action = QAction("Open session", self)
        open_action.setShortcut("Ctrl+O")
        open_action.setIcon(QIcon.fromTheme("document-open"))
        open_action.triggered.connect(self.open_file_dialog)
        session_menu.addAction(open_action)

        self.reset_action = QAction("Reset session...", self)
        self.reset_action.setShortcut("Ctrl+R")
        self.reset_action.setEnabled(False)
        self.reset_action.setIcon(get_icon("refresh"))
        self.reset_action.triggered.connect(self.reset_session)
        session_menu.addAction(self.reset_action)

        self.save_action = QAction("Save session", self)
        self.save_action.setShortcut("Ctrl+S")
        self.save_action.setEnabled(False)
        self.save_action.setIcon(QIcon.fromTheme("document-save"))
        self.save_action.triggered.connect(self.save_session)
        session_menu.addAction(self.save_action)

        drawing_flags = self.menuBar().addMenu("Draw")

        self.view_labels = QAction("Labels", self)
        self.view_labels.setShortcut("Alt+L")
        self.view_contours = QAction("Contours", self)
        self.view_contours.setShortcut("Alt+C")
        self.view_centroids = QAction("Centroids", self)
        self.view_centroids.setShortcut("Alt+P")
        self.view_bboxes = QAction("Bounding boxes", self)
        self.view_bboxes.setShortcut("Alt+B")
        self.view_trails = QAction("Trails", self)
        self.view_trails.setShortcut("Alt+T")
        self.view_ROIs = QAction("Regions of interest", self)
        self.view_ROIs.setShortcut("Alt+R")

        drawing_flags.addActions(
            (
                self.view_labels,
                self.view_contours,
                self.view_centroids,
                self.view_bboxes,
                self.view_trails,
                self.view_ROIs,
            )
        )

        for action in drawing_flags.actions():
            action.setCheckable(True)
            action.toggled.connect(self.video_player.update)

        find_identity_action = QAction("Find identity", self)
        find_identity_action.setIcon(QIcon.fromTheme("edit-find"))
        find_identity_action.setShortcut("Ctrl+F")
        find_identity_action.triggered.connect(self.find_identity)

        drawing_flags.addSeparator()
        drawing_flags.addAction(find_identity_action)

        # Defaults
        self.view_labels.setChecked(True)
        self.view_contours.setChecked(True)
        self.view_centroids.setChecked(True)
        self.view_bboxes.setChecked(False)
        self.view_trails.setChecked(True)
        self.view_ROIs.setChecked(False)

        tooltips = toml.load(Path(__file__).parent / "tooltips.toml")

        self.additional_info.metadata_visibility.setToolTip(
            tooltips["metadata_visibility"]
        )
        self.interpolator.apply_btn.setToolTip(tooltips["apply_interpolation"])
        self.interpolator.abort_btn.setToolTip(tooltips["abort_interpolation"])
        self.errorsExplorer.jumps_th.setToolTip(tooltips["jumps_th"])
        self.errorsExplorer.reset_jumps.setToolTip(tooltips["reset_jumps"])
        self.errorsExplorer.jumps_th_label.setToolTip(tooltips["jumps_th"])
        self.errorsExplorer.jumps_th.setToolTip(tooltips["jumps_th"])
        self.errorsExplorer.update_btn.setToolTip(tooltips["update_errors"])
        self.errorsExplorer.autoselect_errors.setToolTip(tooltips["autoselect_error"])
        self.respect_tracking_intervals.setToolTip(tooltips["tracking_intervals_check"])
        self.interpolator.interpolation_order_box.setToolTip(
            tooltips["interpolation_order"]
        )
        self.interpolator.interpolation_order_label.setToolTip(
            tooltips["interpolation_order"]
        )
        for index in range(self.interpolator.input_size_row.count()):
            self.interpolator.input_size_row.itemAt(index).widget().setToolTip(
                tooltips["input_size"]
            )

        self.unsaved_changes = False

    def open_file_dialog(self) -> None:
        file_dialog = QFileDialog()
        settings_key = f"{self.__class__.__name__}_filedialog_state"
        state = self.settings.value(settings_key, b"")
        if state:
            file_dialog.restoreState(state)
        file_paths = file_dialog.getExistingDirectory(
            self, "Open session directory", options=QFileDialog.Option.ShowDirsOnly
        )
        self.settings.setValue(settings_key, file_dialog.saveState())
        self.open_session(file_paths)

    def find_identity(self) -> None:
        """Displays a QInputDialog to select an identity to, then, find
        its blob, select it and center the video canvas to its centroid"""
        to_find, success = QInputDialog.getText(
            self,
            "",
            "Identity to find:",
            text=str(self.selected_id) if self.selected_id not in (None, -1) else "",
            flags=Qt.WindowType.SplashScreen,
        )
        to_find = to_find.strip()

        if not success or not to_find:
            return

        if to_find.isdigit():
            identity_to_find = int(to_find)
        else:
            try:
                identity_to_find = self.id_labels.labels.index(to_find)
            except ValueError:
                QMessageBox.warning(
                    self, "Find error", f"Identity not recognized: {to_find!r}"
                )
                return

        for blob in self.blobs.blobs_in_video[self.current_frame_number]:
            for identity, centroid in blob.final_ids_and_centroids:
                if identity == identity_to_find:
                    self.selected_blob = blob
                    self.selected_id = identity
                    self.selection_last_location = centroid
                    self.current_frame_number = -1  # this makes info_widget to update
                    self.video_player.center_canvas_at(*centroid, self.max_zoom)
                    return

        QMessageBox.warning(
            self, "Find error", f"Identity {identity_to_find} not found in this frame"
        )

    def keyPressEvent(self, a0: QKeyEvent) -> None:
        if a0.key() in (Qt.Key.Key_Return, Qt.Key.Key_Enter):
            self.id_groups.uncheck_edit_buttons()
            self.setup_points.add.setChecked(False)

    def manageDropedPaths(self, paths: Sequence[str]) -> None:
        if not paths:
            return
        if len(paths) > 1:
            QMessageBox.warning(
                self,
                "Error with dragged files",
                "Cannot open more than one session at once",
            )
            return
        self.open_session(paths[0])

    def go_to_error(
        self, kind: str, start: int, length: int, where: np.ndarray, identity: int
    ) -> None:
        if where.ndim == 2:
            # Set the zoom to capture all positions of 'where'
            xmax, ymax = np.nanmax(where, axis=0)
            xmin, ymin = np.nanmin(where, axis=0)
            zoom_scale = max(self.max_zoom, 1.8 * (xmax - xmin), 1.8 * (ymax - ymin))
            self.video_player.center_canvas_at(
                0.5 * (xmax + xmin), 0.5 * (ymin + ymax), zoom_scale=zoom_scale
            )
            where = where[0]
        else:
            self.video_player.center_canvas_at(*where, zoom_scale=self.max_zoom)
        self.selection_last_location = None if np.isnan(where).any() else tuple(where)  # type: ignore

        self.selected_id = identity
        if kind in ("Jump", "Miss id"):
            self.interpolator.set_interpolation_params(identity, start, start + length)
        else:
            self.interpolator.setActivated(False)
        self.video_player.setCurrentFrame(
            start - 1 if start > 0 and kind in ("Jump", "Miss id") else start, True
        )

    def reset_session(self) -> None:
        start, finish = self.reset_session_dialog.exec()
        logging.info(f"Reseting session from {start} to {finish}")
        if start is None:
            return

        for blobs_in_frame in track(
            self.blobs.blobs_in_video[start:finish], "Resetting user corrections"
        ):
            for blob in blobs_in_frame:
                if blob.added_by_user:
                    self.blobs.blobs_in_video[blob.frame_number].remove(blob)
                else:
                    blob.user_generated_identities = None  # type: ignore
                    blob.user_generated_centroids = None  # type: ignore

        self.errorsExplorer.non_accepted_jumps[start:finish] = True
        self.update_trajectories_range(start, finish)

    def save_session(self) -> None:
        self.session.identities_labels = self.id_labels.get_labels()[1:]
        self.session.identities_colors = [
            c.name() for c in self.id_labels.get_colors()[0][1:]
        ]
        self.session.identities_groups = self.id_groups.get_groups()
        self.session.setup_points = self.setup_points.get_points()
        self.session.length_calibrations = self.length_calibrator.get_calibrations()
        self.session.last_validated = datetime.now()

        saving_thread = SaveSessionObjects(
            self.session, self.blobs, self, self.blobs_path
        )
        progress_bar = QProgressDialog(
            "Saving session, please wait...",
            "Close app",
            0,
            0,
            self,
            Qt.WindowType.SplashScreen,
        )
        progress_bar.canceled.connect(saving_thread.terminate)
        progress_bar.canceled.connect(sys.exit)
        progress_bar.setModal(True)

        saving_thread.finished.connect(self._start_save_trajectories)
        saving_thread.finished.connect(progress_bar.cancel)
        saving_thread.start()
        progress_bar.show()

    def _start_save_trajectories(self):
        progress = QProgressDialog(
            "Computing trajectories",
            "Abort",
            0,
            self.session.number_of_frames + 1,
            self,
            Qt.WindowType.SplashScreen,
        )
        progress.setMinimumDuration(1500)
        progress.setModal(True)

        self.save_thread = SaveTrajectoriesThread(
            self.blobs.blobs_in_video, self.session, self.fragments
        )
        progress.canceled.connect(self.save_thread.quit)
        self.save_thread.finished.connect(self.finish_saving)
        self.save_thread.progress_changed.connect(progress.setValue)
        self.save_thread.start()

    def finish_saving(self) -> None:
        if self.save_thread.success:
            self.unsaved_changes = False

    def check_unsaved_changes(self) -> None | QMessageBox.StandardButton:
        if not self.unsaved_changes:
            return None
        return QMessageBox.question(
            self,
            "Save changes?",
            "There are unsaved changes. Changes which are not saved will be"
            " permanently lost.",
            QMessageBox.StandardButton.Cancel
            | QMessageBox.StandardButton.Discard
            | QMessageBox.StandardButton.Save,
        )

    def ask_for_session(self) -> None:
        answer = (
            QMessageBox.information(
                self,
                "Open session",
                "Please, open a session to validate",
                QMessageBox.StandardButton.Open | QMessageBox.StandardButton.Cancel,
            )
            is QMessageBox.StandardButton.Open
        )
        if answer is QMessageBox.StandardButton.Open:
            self.open_file_dialog()

    def open_session(self, session: Session | Path | str | None) -> None:
        if not session:
            return

        match self.check_unsaved_changes():
            case QMessageBox.StandardButton.Cancel:
                return
            case QMessageBox.StandardButton.Save:
                self.save_session()
            case QMessageBox.StandardButton.Discard:
                pass  # override unsaved changes
            case None:
                pass  # there are not unsaved changes
            case other:
                raise ValueError(other)

        if not isinstance(session, Session):
            session = open_session(self, self.settings, session)
            if session is None:
                return
        self.session = session

        if (
            hasattr(session, "timers")
            and "Tracking session" in session.timers
            and not session.timers["Tracking session"].finished
        ):
            answer = QMessageBox.question(
                self,
                "Loading session warning",
                f"The session you are trying to load ({session}) has not finished,"
                " unexpected behavior can happen. Do you want to continue?",
                QMessageBox.StandardButton.Cancel | QMessageBox.StandardButton.Ok,
            )
            if answer != QMessageBox.StandardButton.Ok:
                return

        loading_thread = LoadSessionObjects(session, self)
        progress_bar = QProgressDialog(
            "Loading session, please wait...",
            "Close app",
            0,
            0,
            self,
            Qt.WindowType.SplashScreen,
        )
        progress_bar.canceled.connect(loading_thread.terminate)
        progress_bar.canceled.connect(sys.exit)
        progress_bar.setModal(True)

        loading_thread.finished.connect(lambda: self._finalize_loading(loading_thread))
        loading_thread.finished.connect(progress_bar.cancel)
        loading_thread.start()
        progress_bar.show()

    def _finalize_loading(self, loading_thread: LoadSessionObjects) -> None:
        if loading_thread.blobs is None:
            QMessageBox.warning(
                self, "Loading session error", "List of blobs not found"
            )
            return
        session = self.session

        self.blobs_path = loading_thread.loaded_from

        # remove selection
        self.selected_blob = None
        self.selected_id = None
        self.selection_last_location = None

        self.id_groups.load_groups(session.identities_groups)
        self.id_labels.load(
            session.identities_labels or [str(i + 1) for i in range(session.n_animals)],
            session.identities_colors,
        )
        self.blobs = loading_thread.blobs
        self.fragments = loading_thread.fragments

        self.additional_info.fragments = self.fragments

        self.video_player.update_video_paths(
            session.video_paths,
            session.number_of_frames,
            (session.width, session.height),
            session.frames_per_second,
        )
        self.n_animals = session.n_animals
        self.n_frames = session.number_of_frames
        self.generate_trajectories(self.blobs.blobs_in_video)
        try:
            self.max_zoom = 2 * session.median_body_length
        except AttributeError:
            logging.warning('No "median_body_length" found in session')
            self.max_zoom = 50 * np.nanmedian(
                np.sqrt((np.diff(self.trajectories, axis=0) ** 2).sum(-1))
            )
        self.centralWidget().setEnabled(True)
        self.dbl_click_dialog = DblClickDialog(self, session.n_animals)

        self.respect_tracking_intervals.setVisible(
            not (
                session.tracking_intervals is None
                or (session.tracking_intervals[0] == [0, session.number_of_frames])
            )
        )

        self.setup_points.load_points(session.setup_points)
        self.length_calibrator.load(
            session.length_calibrations
            if hasattr(session, "length_calibrations")
            else None
        )

        tracking_intervals = (
            session.tracking_intervals
            if self.respect_tracking_intervals.isChecked()
            else None
        )
        self.errorsExplorer.set_references(
            self.trajectories,
            self.unidentified,
            self.duplicated,
            self.blobs,
            tracking_intervals,
        )
        self.interpolator.set_references(
            self.trajectories,
            self.unidentified,
            self.duplicated,
            self.blobs,
            tracking_intervals,
        )
        self.video_player.update()
        self.unsaved_changes = False

        if hasattr(session, "roi_list") and session.roi_list:
            self.view_ROIs.setEnabled(True)
            self.view_ROIs.setChecked(True)
            self.ROI_pathces = build_ROI_patches_from_list(
                session.roi_list, session.width, session.height
            )
        else:
            self.view_ROIs.setChecked(False)
            self.view_ROIs.setEnabled(False)

        self.setWindowTitle("Validator | " + session.session_folder.name)

        self.save_action.setEnabled(True)
        self.reset_action.setEnabled(True)

        self.reset_session_dialog = ResetSessionDialog(self, session.number_of_frames)

    def respect_tracking_intervals_changed(self, checked: bool) -> None:
        "Transfer the status information of respect_tracking_intervals to the widgets that need it"
        tracking_intervals = self.session.tracking_intervals if checked else None
        self.errorsExplorer.set_references(
            self.trajectories,
            self.unidentified,
            self.duplicated,
            self.blobs,
            tracking_intervals,
        )
        self.interpolator.set_references(
            self.trajectories,
            self.unidentified,
            self.duplicated,
            self.blobs,
            tracking_intervals,
        )

    def click_on_canvas(self, event: CanvasMouseEvent) -> None:
        self.selected_blob, self.selected_id, self.selection_last_location = clicked_id(
            self.blobs.blobs_in_video[self.current_frame_number], event
        )
        self.id_groups.selected_id(self.selected_id)
        self.current_frame_number = -1  # this makes info_widget to update
        self.video_player.update()

    def double_click_on_canvas(self, event: CanvasMouseEvent) -> None:
        if (
            self.selected_blob is None
            or self.id_groups.is_active()
            or event.button != Qt.MouseButton.LeftButton
        ):
            return
        self.selected_blob, self.selected_id, self.selection_last_location = clicked_id(
            self.blobs.blobs_in_video[self.current_frame_number], event
        )
        if self.selection_last_location is None:
            # clicked on a blob without centroids
            answer, new_id, propagate = self.dbl_click_dialog.exec_with_description(0)
            if answer == DblClickDialog.Answers.ChangeId:
                self.selected_blob.add_centroid(event.xy_data, new_id)
                self.update_trajectories_range(self.current_frame_number)
            return

        # clicked on a blob with centroid
        answer, new_id, propagate = self.dbl_click_dialog.exec_with_description(
            self.interpolator.animal_id + 1
            if self.interpolator.isEnabled()
            else self.selected_id
        )

        if answer == DblClickDialog.Answers.Reset:
            new_id = None
            answer = DblClickDialog.Answers.ChangeId

        if answer == DblClickDialog.Answers.Remove:
            new_id = -1
            answer = DblClickDialog.Answers.ChangeId

        if answer == DblClickDialog.Answers.ChangeId:
            self.selected_blob.update_identity(
                self.selected_id, new_id, self.selection_last_location
            )
            # If we are interpolating, do not update list of errors, let the interpolation finish
            update_errors = not self.interpolator.isEnabled()
            if propagate:
                lower, upper = self.selected_blob.propagate_identity(
                    self.selected_id, new_id, self.selection_last_location
                )
                if lower != upper:
                    self.light_popup.info(
                        "Identification change",
                        f"Identification propagated from frame {lower} to frame"
                        f" {upper}",
                    )
                self.update_trajectories_range(
                    lower, upper + 1, update_errors=update_errors
                )
            else:
                self.update_trajectories_range(
                    self.current_frame_number, update_errors=update_errors
                )
            return
        if (
            answer == DblClickDialog.Answers.Interpolate
            and self.selected_id is not None
            and self.selected_id > 0
        ):
            self.interpolator.set_interpolation_params(
                self.selected_id,
                self.current_frame_number,
                self.current_frame_number + 1,
            )

    def paint(self, painter: CanvasPainter, frame_number: int) -> None:
        blobs_in_frame = self.blobs.blobs_in_video[frame_number]
        if self.id_groups.is_active():
            cmap, cmap_alpha = self.id_groups.get_cmaps(self.session.n_animals)
        else:
            cmap, cmap_alpha = self.id_labels.get_colors()
        update_info_widget = frame_number != self.current_frame_number
        self.current_frame_number = frame_number

        if self.selected_blob is not None and self.selected_blob not in blobs_in_frame:
            self.selected_blob, self.selection_last_location = find_selected_blob(
                blobs_in_frame, self.selected_id, self.selection_last_location
            )

        if self.view_trails.isChecked():
            paintTrails(self.current_frame_number, painter, self.trajectories, cmap)

        if self.view_ROIs.isChecked():
            painter.setPen(Qt.PenStyle.NoPen)
            painter.setBrush(QColor(255, 0, 0, 50))
            painter.drawPath(self.ROI_pathces)

        paintBlobs(
            self.view_contours.isChecked(),
            self.view_centroids.isChecked(),
            self.view_bboxes.isChecked(),
            self.view_labels.isChecked(),
            painter,
            blobs_in_frame,
            cmap,
            cmap_alpha,
            self.selected_blob,
            self.selection_last_location,
            self.id_labels.get_labels(),
            self.mark_metadata(blobs_in_frame, self.fragments),
        )

        if self.setup_points.isVisible():
            self.setup_points.paint_on_canvas(painter)

        if self.length_calibrator.isVisible():
            self.length_calibrator.paint_on_canvas(painter)

        if self.interpolator.isEnabled():
            self.interpolator.paint_on_canvas(painter, frame_number)

        if update_info_widget:
            self.additional_info.set_data(self.selected_blob)

    def closeEvent(self, event: QCloseEvent) -> None:
        match self.check_unsaved_changes():
            case QMessageBox.StandardButton.Discard:
                return super().closeEvent(event)
            case QMessageBox.StandardButton.Save:
                self.save_session()
                return super().closeEvent(event)
            case QMessageBox.StandardButton.Cancel:
                return event.ignore()
            case None:  # there are not unsaved changes
                return super().closeEvent(event)
            case other:
                raise ValueError(other)

    def update_trajectories_range(
        self, start: int, finish: int | None = None, update_errors: bool = True
    ) -> None:
        finish = start + 1 if finish is None else finish
        logging.debug(f"Updating trajectories from frame {start} to {finish}")
        ids_in_frame = set()
        self.trajectories[start:finish] = np.nan
        self.duplicated[start:finish] = False
        self.unidentified[start:finish] = False
        for frame in range(start, finish):
            ids_in_frame.clear()
            for blob in self.blobs.blobs_in_video[frame]:
                for identity, centroid in blob.final_ids_and_centroids:
                    if identity not in (None, 0):
                        self.trajectories[frame, identity - 1] = centroid
                        if identity in ids_in_frame:
                            self.duplicated[frame, identity - 1] = True
                        ids_in_frame.add(identity)
                    else:
                        self.unidentified[frame] = True
        self.interpolator.trajectories_have_been_updated()
        if update_errors:
            self.errorsExplorer.update_list_of_errors()
        self.video_player.update()
        self.unsaved_changes = True

    def generate_trajectories(self, blobs_in_video: list[list[Blob]]) -> None:
        number_of_frames = len(blobs_in_video)
        self.trajectories = np.full((number_of_frames, self.n_animals, 2), np.nan)
        self.unidentified = np.zeros(number_of_frames, bool)
        self.duplicated = np.zeros((number_of_frames, self.n_animals), bool)
        ids_in_frame: set[int] = set()

        progress_bar = QProgressDialog(
            "Analyzing trajectories",
            "Close app",
            0,
            number_of_frames - 1,
            self,
            Qt.WindowType.SplashScreen,
        )
        progress_bar.setMinimumDuration(1000)
        progress_bar.canceled.connect(sys.exit)
        progress_bar.setModal(True)

        for frame, blobs_in_frame in enumerate(blobs_in_video):
            progress_bar.setValue(frame)
            ids_in_frame.clear()
            for blob in blobs_in_frame:
                for identity, centroid in blob.final_ids_and_centroids:
                    if identity not in (None, 0):
                        self.trajectories[frame, identity - 1] = centroid
                        if identity in ids_in_frame:
                            self.duplicated[frame, identity - 1] = True
                        ids_in_frame.add(identity)
                    else:
                        self.unidentified[frame] = True


def clicked_id(
    blobs: list[Blob], click: CanvasMouseEvent
) -> tuple[Blob | None, int | None, tuple[float, float] | None]:
    candidates: list[tuple[Blob, int | None, tuple[float, float], float]] = []

    for blob in blobs:
        if blob.contains_point(click.xy_data):
            for identity, centroid in blob.final_ids_and_centroids:
                dist = click.sq_distance_to(centroid)
                candidates.append((blob, identity, centroid, dist))
            if not candidates:  # blob with no centroids
                return blob, -1, None
            break

    if candidates:
        return min(candidates, key=lambda x: x[-1])[:-1]

    for blob in blobs:
        for identity, centroid in blob.final_ids_and_centroids:
            dist = click.sq_distance_to(centroid)
            if dist < (SELECT_POINT_DIST * click.zoom):
                candidates.append((blob, identity, centroid, dist))

    if candidates:
        return min(candidates, key=lambda x: x[-1])[:-1]

    return None, -1, None


class SaveTrajectoriesThread(QThread):
    progress_changed = Signal(int)

    def __init__(
        self,
        blobs_in_video: list[list[Blob]],
        session: Session,
        list_of_fragments: list[Fragment] | None,
    ) -> None:
        super().__init__()
        self.blobs_in_video = blobs_in_video
        self.fragments = list_of_fragments
        self.session = session
        self.success = False
        self.finished.connect(
            lambda: self.progress_changed.emit(len(self.blobs_in_video) + 1)
        )

    def run(self) -> None:
        self.abort = False

        trajectories = produce_output_dict(
            self.blobs_in_video,
            self.session,
            progress_bar=self.progress_changed,
            abort=lambda: self.abort,
        )
        if self.abort:
            return
        if not trajectories:  # empty dict
            logging.error("Trajectories could not be produced")
            return

        save_trajectories(
            self.session.trajectories_folder,
            trajectories,
            self.session.trajectories_formats,
        )

        self.progress_changed.emit(self.session.number_of_frames)
        self.success = True

    def quit(self) -> None:
        self.abort = True


class ResetSessionDialog(QDialog):
    """Pop up to select the range of the user corrections reset.
    Reset is activated in the menu bar / Session / Reset session"""

    class Answers(Enum):
        Cancel = 0
        RangeReset = 1
        AllReset = 2

    def __init__(self, parent, n_frames: int) -> None:
        super().__init__(parent)

        layout = QVBoxLayout()
        self.setLayout(layout)
        layout.addWidget(
            QLabel(
                '<span style="font-weight:600">Resetting removes all user corrections'
            ),
            alignment=Qt.AlignmentFlag.AlignCenter,
        )
        layout.addWidget(
            QLabel("Range of frames to reset:"), alignment=Qt.AlignmentFlag.AlignCenter
        )
        self.double_slider = LabelRangeSlider(0, n_frames, parent)
        layout.addWidget(self.double_slider)
        btn_layout = QHBoxLayout()
        cancel_btn = QPushButton(get_icon("cancel"), "Cancel")
        range_btn = QPushButton(get_icon("refresh"), "Reset range")
        all_btn = QPushButton(get_icon("refresh"), "Reset all")
        btn_layout.addWidget(cancel_btn)
        btn_layout.addWidget(range_btn)
        btn_layout.addWidget(all_btn)
        layout.addLayout(btn_layout)

        cancel_btn.clicked.connect(lambda: self.done(self.Answers.Cancel.value))
        range_btn.clicked.connect(lambda: self.done(self.Answers.RangeReset.value))
        all_btn.clicked.connect(lambda: self.done(self.Answers.AllReset.value))

        cancel_btn.setAutoDefault(False)
        range_btn.setAutoDefault(False)
        all_btn.setAutoDefault(False)

    def exec(self) -> tuple[int | None, int | None]:  # type: ignore
        match self.Answers(super().exec()):
            case self.Answers.Cancel:
                return None, None
            case self.Answers.RangeReset:
                return self.double_slider.value()
            case self.Answers.AllReset:
                return self.double_slider.minimum(), self.double_slider.maximum()



import logging

import numpy as np

from idtrackerai import ListOfBlobs, conf
from idtrackerai.utils import track


class ModelArea:
    """Model of the area used to perform a first discrimination between blobs
    representing single individual and multiple touching animals (crossings)

    Attributes
    ----------

    median : float
        median of the area of the blobs segmented from portions of the video in
        which all the animals are visible (not touching)
    mean : float
        mean of the area of the blobs segmented from portions of the video in
        which all the animals are visible (not touching)
    std : float
        standard deviation of the area of the blobs segmented from portions of
        the video in which all the animals are visible (not touching)
    std_tolerance : int
        tolerance factor

    Methods
    -------
    __call__:
      some description
    """

    def __init__(self, list_of_blobs: ListOfBlobs, number_of_animals: int):
        """computes the median and standard deviation of the area of all the blobs
        in the the video and the median of the the diagonal of the bounding box.
        """
        # areas are collected throughout the entire video inthe cores of the
        # global fragments
        logging.info(
            "Initializing ModelArea for individual/crossing blob initial classification"
        )
        areas = []
        if number_of_animals > 0:
            for blobs_in_frame in list_of_blobs.blobs_in_video:
                if len(blobs_in_frame) == number_of_animals:
                    areas += (blob.area for blob in blobs_in_frame)

        if not areas:
            # either n_animals is 0 or there are no global fragments
            areas = [b.area for b in list_of_blobs.all_blobs]

        areas = np.asarray(areas)
        self.median = np.median(areas)
        self.mean = areas.mean()
        self.std = areas.std()
        self.std_tolerance = conf.MODEL_AREA_SD_TOLERANCE
        self.tolerance = self.std_tolerance * self.std
        logging.info(
            f"Model area computed with {len(areas)} blobs: "
            f"mean={self.mean:.1f}, median={self.median:.1f}, "
            f"and std={self.std:.1f} pixels"
        )

    def __call__(self, area) -> bool:
        return (area - self.median) < self.tolerance


def compute_body_length(list_of_blobs: ListOfBlobs, number_of_animals: int) -> float:
    """computes the median of the the diagonal of the bounding box."""
    # areas are collected throughout the entire video in the cores of
    # the global fragments
    body_lengths = []
    if number_of_animals > 0:
        for blobs_in_frame in track(
            list_of_blobs.blobs_in_video, "Computing body lengths"
        ):
            if len(blobs_in_frame) == number_of_animals:
                body_lengths += (blob.extension for blob in blobs_in_frame)

    if not body_lengths:
        # either n_animals is 0 or there are no global fragments
        body_lengths = [
            b.extension
            for b in track(list_of_blobs.all_blobs, "Computing body lengths")
        ]

    median = np.median(body_lengths)
    logging.info(f"The median body length is {median:.1f} pixels")
    return float(median)
    # return np.percentile(body_lengths, 80)



from collections.abc import Sequence
from pathlib import Path

import cv2
import numpy as np
from qtpy.QtCore import QRect, Qt, QUrl
from qtpy.QtGui import QDesktopServices, QIcon, QImage, QPixmap
from qtpy.QtWidgets import (
    QApplication,
    QFileDialog,
    QHBoxLayout,
    QLabel,
    QMessageBox,
    QProgressDialog,
    QPushButton,
    QSpinBox,
    QSplitter,
    QVBoxLayout,
    QWidget,
)
from superqt import QToggleSwitch

from idtrackerai.GUI_tools import (
    CanvasPainter,
    GUIBase,
    IdLabels,
    LabelRangeSlider,
    QHLine,
    VideoPlayer,
    open_session,
)
from idtrackerai.utils import load_trajectories

from .general_video import draw_general_frame, generate_general_video
from .individual_videos import (
    draw_collage_frame,
    generate_individual_video,
    get_geometries,
    get_individual_miniframes,
)


class VideoGeneratorGUI(GUIBase):

    def __init__(self) -> None:
        super().__init__()

        self.setWindowTitle("Video Generator")
        self.documentation_url = (
            "https://idtracker.ai/latest/user_guide/video_generators.html"
        )

        self.video_player = VideoPlayer(self)
        self.video_player.is_muted = True
        self.video_player.draw_in_color.setEnabled(False)
        self.video_player.draw_in_color.setChecked(True)
        self.widgets_to_close.append(self.video_player)
        self.video_player.painting_time.connect(self.preview_frame)

        self.open_button = QPushButton()
        self.open_button.setIcon(QIcon.fromTheme("document-open"))
        self.open_button.setText("Open Session")
        self.open_button.setShortcut("Ctrl+O")
        self.open_button.setToolTip("Open a session to generate videos")
        self.open_button.clicked.connect(self.open)

        video_kind_row = QWidget()
        video_kind_layout = QHBoxLayout()
        video_kind_row.setLayout(video_kind_layout)
        video_kind_layout.setContentsMargins(0, 0, 0, 0)
        video_kind_layout.addWidget(QLabel("General Video"))
        video_kind_layout.setAlignment(Qt.AlignmentFlag.AlignCenter)
        self.individual_kind_switch = QToggleSwitch(self)
        self.individual_kind_switch.setChecked(False)
        self.individual_kind_switch.setToolTip(
            "Switch between General Video and Individual Videos"
        )
        self.individual_kind_switch.toggled.connect(self.toggle_video_kind)
        video_kind_layout.addWidget(self.individual_kind_switch)
        video_kind_layout.addWidget(QLabel("Individual Video"))

        self.id_labels = IdLabels()
        self.id_labels.needToDraw.connect(self.video_player.update)

        self.individual_size_row = QWidget(self)
        individual_size_layout = QHBoxLayout()
        self.individual_size_row.setLayout(individual_size_layout)
        individual_size_layout.setContentsMargins(0, 0, 0, 0)
        self.individual_size = QSpinBox(self)
        self.individual_size.setSuffix(" pixels")
        self.individual_size.setRange(2, 1000)
        self.individual_size.setSingleStep(2)
        self.individual_size.valueChanged.connect(self.individual_size_changed)
        individual_size_layout.addWidget(QLabel("Individual Size:"))
        individual_size_layout.addWidget(self.individual_size)
        self.individual_size_row.setVisible(False)

        render_in_gray_row = QWidget(self)
        render_in_gray_layout = QHBoxLayout()
        render_in_gray_row.setLayout(render_in_gray_layout)
        render_in_gray_layout.setContentsMargins(0, 0, 0, 0)
        self.render_in_color = QToggleSwitch()
        self.render_in_color.setChecked(True)
        self.render_in_color.setToolTip(
            "Render the video in color (checked) or grayscale (unchecked)"
        )
        self.render_in_color.toggled.connect(self.video_player.update)
        render_in_gray_layout.addWidget(QLabel("Render in Color:"))
        render_in_gray_layout.addWidget(self.render_in_color)

        self.general_trace_length_row = QWidget(self)
        general_trace_length_layout = QHBoxLayout()
        self.general_trace_length_row.setLayout(general_trace_length_layout)
        general_trace_length_layout.setContentsMargins(0, 0, 0, 0)
        self.general_trace_length = QSpinBox(self)
        self.general_trace_length.setSuffix(" frames")
        self.general_trace_length.setRange(0, 1000)
        self.general_trace_length.setValue(20)
        self.general_trace_length.valueChanged.connect(self.video_player.update)
        general_trace_length_layout.addWidget(QLabel("General Trace Length:"))
        general_trace_length_layout.addWidget(self.general_trace_length)

        self.resize_factor_row = QWidget(self)
        resize_factor_layout = QHBoxLayout()
        self.resize_factor_row.setLayout(resize_factor_layout)
        resize_factor_layout.setContentsMargins(0, 0, 0, 0)
        resize_factor_layout.addWidget(QLabel("Output resolution:"))
        self.resize_factor = QSpinBox(self)
        self.resize_factor.setSuffix(" %")
        self.resize_factor.setRange(10, 100)
        self.resize_factor.setValue(100)
        self.resize_factor.setSingleStep(10)
        self.resize_factor.valueChanged.connect(self.resize_factor_changed)
        resize_factor_layout.addWidget(self.resize_factor)

        self.interval = LabelRangeSlider(0, 1, self)
        self.interval.single_value_changed.connect(self.video_player.setCurrentFrame)

        self.generate_video_button = QPushButton(self)
        self.generate_video_button.setIcon(QIcon.fromTheme("document-save"))
        self.generate_video_button.setText("Generate and save video")
        self.generate_video_button.setToolTip(
            "Generate the video with the current settings"
        )
        self.generate_video_button.clicked.connect(self.generate_video)

        layout = self.centralWidget().layout()
        assert layout is not None

        self.controls_widget = QWidget(self)
        controls = QVBoxLayout()
        self.controls_widget.setLayout(controls)
        controls.addWidget(video_kind_row, alignment=Qt.AlignmentFlag.AlignHCenter)
        controls.addWidget(QHLine())
        identity_labels_switch = QToggleSwitch("Draw identity labels")
        identity_labels_switch.setChecked(True)
        identity_labels_switch.toggled.connect(self.id_labels.set_labels_enabled)
        identity_labels_switch.toggled.connect(self.video_player.update)
        controls.addWidget(identity_labels_switch)
        controls.addWidget(self.id_labels)
        controls.addWidget(QHLine())
        controls.addWidget(self.general_trace_length_row)
        controls.addWidget(self.individual_size_row)
        controls.addWidget(self.resize_factor_row)
        controls.addWidget(render_in_gray_row)
        controls.addWidget(QHLine())
        controls.addWidget(QLabel("Video generation time-range:"))
        controls.addWidget(self.interval)
        controls.addWidget(self.generate_video_button)

        left_widget = QWidget(self)
        left_layout = QVBoxLayout()
        left_layout.setAlignment(Qt.AlignmentFlag.AlignVCenter)
        left_widget.setLayout(left_layout)
        left_layout.setContentsMargins(0, 0, 0, 0)
        left_layout.addWidget(self.open_button)
        left_layout.addWidget(self.controls_widget)

        main_splitter = QSplitter(Qt.Orientation.Horizontal)
        main_splitter.addWidget(left_widget)
        main_splitter.addWidget(self.video_player)
        main_splitter.setStretchFactor(0, 1)
        main_splitter.setStretchFactor(1, 3)
        layout.addWidget(main_splitter)

        self.video_player.setEnabled(False)
        self.controls_widget.setEnabled(False)

    def closeEvent(self, event):
        if (
            hasattr(self, "session")
            and self.session is not None
            and self.id_labels.isEnabled()
        ):
            self.session.identities_labels = self.id_labels.get_labels()[1:]
            self.session.identities_colors = [
                c.name() for c in self.id_labels.get_colors()[0][1:]
            ]
            self.session.save()
        super().closeEvent(event)

    def generate_video(self):
        if self.individual_kind_switch.isChecked():
            output_path = QFileDialog.getExistingDirectory(
                self, "Select Output Folder", str(self.session.session_folder)
            )
        else:
            output_path = QFileDialog.getSaveFileName(
                self,
                "Save Video",
                str(
                    self.session.session_folder
                    / (self.session.video_paths[0].stem + "_tracked.avi")
                ),
                "Video Files (*.avi *.mp4)",
            )[0]

        if not output_path:
            return

        progress_dialog = QProgressDialog("Generating Video", "Cancel", 0, 100, self)
        progress_dialog.setMinimumDuration(0)

        def progress_callback(progress: float):
            progress_dialog.setValue(int(progress * 100))
            QApplication.processEvents()
            if progress_dialog.wasCanceled():
                raise KeyboardInterrupt("Video generation cancelled by user.")

        try:
            if self.individual_kind_switch.isChecked():
                generate_individual_video(
                    self.session,
                    self.session.trajectories_folder,
                    starting_frame=self.interval.value()[0],
                    ending_frame=self.interval.value()[1],
                    miniframe_size=self.individual_size.value(),
                    labels=(
                        self.id_labels.get_labels()[1:]
                        if self.id_labels.isEnabled()
                        else None
                    ),
                    callback=progress_callback,
                    draw_in_gray=not self.render_in_color.isChecked(),
                    output_dir=output_path,
                )
            else:
                generate_general_video(
                    self.session,
                    self.session.trajectories_folder,
                    starting_frame=self.interval.value()[0],
                    ending_frame=self.interval.value()[1],
                    centroid_trace_length=self.general_trace_length.value(),
                    resize_factor=self.resize_factor.value() / 100,
                    labels=(
                        self.id_labels.get_labels()[1:]
                        if self.id_labels.isEnabled()
                        else None
                    ),
                    callback=progress_callback,
                    output_path=output_path,
                    draw_in_gray=not self.render_in_color.isChecked(),
                    colors=[
                        (c.red(), c.green(), c.blue())
                        for c in self.id_labels.get_colors()[0][1:]
                    ],
                )
        except KeyboardInterrupt:
            pass  # User cancelled the operation
        else:
            msg_box = QMessageBox(self)
            msg_box.setIcon(QMessageBox.Icon.Information)
            msg_box.setWindowTitle("Done")
            msg_box.setText(f"Video generation finished, saved in {output_path}.")
            open_folder_btn = msg_box.addButton(
                "Open File Location", QMessageBox.ButtonRole.ActionRole
            )
            open_folder_btn.setIcon(QIcon.fromTheme("folder-open"))
            msg_box.addButton(QMessageBox.StandardButton.Ok)
            msg_box.exec()

            if msg_box.clickedButton() == open_folder_btn:
                if Path(output_path).is_file():
                    # If output_path is a file, get its directory
                    output_path = str(Path(output_path).parent)
                try:
                    QDesktopServices.openUrl(QUrl.fromLocalFile(output_path))
                except Exception as e:
                    QMessageBox.warning(
                        self, "Error Opening Folder", f"Could not open the folder: {e}"
                    )
        progress_dialog.close()

    def individual_size_changed(self, value: int):
        self.positions, self.individual_output_shape = get_geometries(
            self.session.n_animals, value
        )

        if not self.individual_kind_switch.isChecked():
            return
        self.video_player.canvas.adjust_zoom_to(*self.individual_output_shape[::-1])
        self.video_player.update()

    def toggle_video_kind(self, checked: bool):
        if checked:
            self.video_player.canvas.adjust_zoom_to(*self.individual_output_shape[::-1])
            self.individual_size_row.setVisible(True)
            self.general_trace_length_row.setVisible(False)
            self.resize_factor_row.setVisible(False)
            self.id_labels.set_colors_enabled(False)
        else:
            self.video_player.canvas.adjust_zoom_to(
                self.session.width, self.session.height
            )
            self.individual_size_row.setVisible(False)
            self.general_trace_length_row.setVisible(True)
            self.resize_factor_row.setVisible(True)
            self.id_labels.set_colors_enabled(True)

        self.video_player.update()

    def resize_factor_changed(self, value: int):
        self.resized_trajectories = (self.trajectories * value / 100).astype(int)
        self.video_player.update()

    def manageDropedPaths(self, paths: Sequence[str]) -> None:
        if not paths:
            return
        if len(paths) > 1:
            QMessageBox.warning(
                self,
                "Error with dragged files",
                "Cannot open more than one session at once",
            )
            return
        self.open(paths[0])

    def open(self, session_path: Path | str | None = None):
        session = open_session(self, self.settings, session_path)
        if session is None:
            return
        self.session = session

        try:
            self.trajectories = np.nan_to_num(
                load_trajectories(self.session.session_folder)["trajectories"], nan=-100
            )
        except FileNotFoundError as exc:
            QMessageBox.warning(
                self,
                "Trajectories not found",
                f"{exc}. Did the session finish tracking successfully?",
            )
            return

        self.resize_factor_changed(self.resize_factor.value())

        self.individual_size.setValue(2 * (int(session.median_body_length) // 2))

        self.video_player.update_video_paths(
            session.video_paths,
            session.number_of_frames,
            (session.width, session.height),
            session.frames_per_second,
        )

        self.positions, self.individual_output_shape = get_geometries(
            session.n_animals, self.individual_size.value()
        )

        self.interval.blockSignals(True)
        self.interval.setMaximum(session.number_of_frames - 1)
        self.interval.setValue((0, session.number_of_frames - 1))
        self.interval.blockSignals(False)

        self.id_labels.load(
            session.identities_labels
            or list(map(str, range(1, session.n_animals + 1))),
            session.identities_colors,
        )
        self.toggle_video_kind(self.individual_kind_switch.isChecked())

        self.video_player.setEnabled(True)
        self.controls_widget.setEnabled(True)

    def preview_frame(
        self,
        painter: CanvasPainter,
        frame_index: int,
        original_frame: np.ndarray | None,
    ):
        if original_frame is None or not hasattr(self, "session"):
            return

        assert original_frame.shape[2] == 3
        if not self.render_in_color.isChecked():
            original_frame = cv2.cvtColor(original_frame, cv2.COLOR_BGR2GRAY)

        if self.individual_kind_switch.isChecked():
            miniframes = get_individual_miniframes(
                original_frame,
                (self.trajectories[frame_index]).astype(int),
                self.individual_size.value(),
            )

            frame = draw_collage_frame(
                self.positions,
                miniframes,
                self.individual_output_shape,
                (
                    self.id_labels.get_labels()[1:]
                    if self.id_labels.labels_are_enabled()
                    else None
                ),
            )
            preview_shape = self.video_player.video_drawing_origin
        else:
            resize_factor = self.resize_factor.value() / 100
            if resize_factor != 1:
                original_frame = cv2.resize(
                    original_frame, None, fx=resize_factor, fy=resize_factor
                )
            frame = draw_general_frame(
                original_frame,
                frame_index,
                self.resized_trajectories,
                self.general_trace_length.value(),
                self.id_labels.get_colors()[0][1:],
                self.id_labels.get_labels()[1:] if self.id_labels.isEnabled() else None,
            )
            preview_shape = QRect(0, 0, self.session.width, self.session.height)

        painter.drawPixmap(
            preview_shape,
            QPixmap.fromImage(
                QImage(
                    frame.data,
                    frame.shape[1],
                    frame.shape[0],
                    frame.shape[1] * 3,
                    QImage.Format.Format_BGR888,
                )
            ),
        )



"A simple module to inspect contrastive clusters after tracking, see :ref:`cluster inspection`."

import argparse
import logging
from pathlib import Path

import matplotlib.pyplot as plt
import numpy as np
import torch
from matplotlib.axes import Axes
from scipy.spatial.distance import cdist
from sklearn.manifold import TSNE

from idtrackerai import IdtrackeraiError, ListOfBlobs, ListOfFragments, conf
from idtrackerai.base.network import DEVICE, ResNet18, get_onthefly_dataloader
from idtrackerai.utils import (
    load_trajectories,
    manage_exception,
    resolve_path,
    track,
    wrap_entrypoint,
)

logging.getLogger("matplotlib").setLevel(logging.WARNING)


@torch.inference_mode()
def inspect_clusters(
    session_path: Path, images_per_id: float = 500, gt_path: Path | None = None
) -> None:
    """Use the trained ResNet to embed all images used in contrastive
    training or a subset of them, compute the t-SNE representation in
    2D and save the results in a .png scatter plot and a .csv file.

    Parameters
    ----------
    session_path : Path
        _description_
    images_per_id : int, optional
        _description_, by default 500

    Raises
    ------
    IdtrackeraiError
        _description_
    """
    plt.style.use("dark_background")

    session_path = resolve_path(session_path)
    logging.info(f"Computing t-SNE of {session_path}")

    frags = ListOfFragments.load(
        session_path / "preprocessing/list_of_fragments.json", reconnect=False
    )

    for file in ("identifier_contrastive.model.pt", "contrastive_checkpoint.pt"):
        try:
            resnet = ResNet18.from_file(session_path / "accumulation" / file).to(DEVICE)
        except FileNotFoundError:
            continue
        else:
            resnet.eval()
            break
    else:
        raise IdtrackeraiError(
            f"Could not find the contrastive model weights in the session folder {session_path}"
        )

    save_folder = session_path / "cluster_inspection"
    save_folder.mkdir(exist_ok=True)

    locations = []
    predicted_labels = []
    frames = []
    frag_indices = []
    for frag in frags.individual_fragments:
        if len(frag) >= conf.MIN_N_FRAMES_TO_BE_A_CANDIDATE_FOR_ACCUMULATION:
            locations += frag.image_locations
            predicted_labels += [
                frag.identity if frag.identity is not None else 0
            ] * len(frag)
            frames += range(frag.start_frame, frag.end_frame)
            frag_indices += [frag.identifier] * len(frag)

    selection = np.arange(len(predicted_labels))
    logging.info(f"Collected {len(selection)} images in total")
    if images_per_id > 0 and len(selection) > images_per_id * frags.n_animals:
        images_per_id = int(images_per_id)
        logging.info(f"Using a subset of {images_per_id * frags.n_animals} images")
        np.random.shuffle(selection)
        selection = selection[: images_per_id * frags.n_animals]

    locations = np.asarray(locations)[selection]
    predicted_labels = np.asarray(predicted_labels)[selection]
    frames = np.asarray(frames)[selection]
    frag_indices = np.asarray(frag_indices)[selection]

    # sort images by frame to speed up loading
    order = np.argsort(frames)
    locations = locations[order]
    predicted_labels = predicted_labels[order]
    frames = frames[order]
    frag_indices = frag_indices[order]

    data_loader = get_onthefly_dataloader(locations, frags.id_images_file_paths)

    embs = []
    for images, _labels in track(data_loader, "Embedding images"):
        embs.append(resnet(images.to(DEVICE)).numpy(force=True))
    embs = np.concatenate(embs)

    logging.info(f"Computing t-SNE with {len(embs)} data points")
    tsne = TSNE(n_jobs=4).fit_transform(embs)

    scatter_plot(
        *tsne.T, labels=predicted_labels, file=save_folder / "predicted_t-SNE.png"
    )

    if gt_path is not None:
        # if there are groundtruth trajectories, lets get the labels from there
        groundtruth = load_trajectories(gt_path)["trajectories"]
        # we need the blob's centroids
        blobs = ListOfBlobs.load(session_path / "preprocessing/list_of_blobs.pickle")
        gt_labels = []
        for frame, frag_idx in zip(frames, frag_indices):
            assert isinstance(frame, (np.integer, int))  # for Pylance
            for blob in blobs.blobs_in_video[frame]:
                if blob.fragment_identifier == frag_idx:
                    gt_labels.append(
                        np.argmin(cdist([blob.centroid], groundtruth[frame]))
                    )
                    break
            else:
                raise RuntimeError(f"Blob with {frag_idx=} not found in {frame=}")
        gt_labels = np.asarray(gt_labels) + 1

        scatter_plot(
            *tsne.T, labels=gt_labels, file=save_folder / "groundtruth_t-SNE.png"
        )

    else:
        gt_labels = predicted_labels * 0

    np.savetxt(
        save_folder / "embeddings.csv",
        np.column_stack((frag_indices, embs, tsne, predicted_labels, gt_labels)),
        fmt=["%4d"] + ["%+.4f"] * (len(embs[0]) + 2) + ["%d"] * 2,
        delimiter=", ",
        header="fragment, "
        + ", ".join(f"  dim_{i}" for i in range(8))
        + ",  t-SNE_1,  t-SNE_2"
        + ", predicted_id, groundtruth_id",
        comments="",
    )
    logging.info(f"Plot and csv saved in {save_folder}")


def scatter_plot(x: np.ndarray, y: np.ndarray, labels: np.ndarray, file: Path | str):
    """Expected Numpy array with IDs from 1 to n_animals, 0 values meaning unknown"""
    colormap = plt.get_cmap("hsv")
    colormap.set_bad("white")
    colormap.set_under("white")

    labels = labels.astype(float)
    labels -= 1
    norm = labels.max() + 1
    if norm > 0:
        # hsv colormap is circular, min and max values have the same color
        labels /= labels.max() + 1

    fig = plt.figure(figsize=(8, 8))
    ax = Axes(fig, (0.0, 0.0, 1.0, 1.0))
    ax.set_axis_off()
    fig.add_axes(ax)
    ax.scatter(x, y, c=colormap(labels), s=3, lw=0)
    ax.set(aspect=1)
    fig.savefig(file)


@wrap_entrypoint
def idtrackerai_inspect_clusters_entrypoint():
    argparser = argparse.ArgumentParser(
        description=(
            "Use the trained contrastive network (ResNet) to compute "
            "and store the image embeddings. It then generates a "
            "scatter plot of their t-SNE representation, enabling "
            "visual inspection of the resulting clusters. The results "
            "are saved in 'session_folder/cluster_inspection'"
        )
    )
    argparser.add_argument(
        "session_paths",
        help=(
            "Session paths to check the t-SNE on. "
            "Multiple session paths can be provided."
        ),
        type=Path,
        nargs="+",
    )
    argparser.add_argument(
        "--images_per_id",
        help=(
            "Sets the maximum number of images per animal to sample for speeding "
            "up t-SNE computation. The default value is 500. To disable subsampling, "
            "set this option to 'inf'. Note that subsampling is performed randomly "
            "across all identities, so the exact number of images per class may vary."
        ),
        type=float,
        default=500,
    )
    argparser.add_argument(
        "--gt_path",
        help=(
            "Specifies the path to the ground truth trajectories used to compare "
            "image centroids and extract their ground truth identities. These "
            "identities are used to assign colors in the groundtruth scatter plot and populate "
            "the groundtruth_id column in the CSV output. The path can point to a "
            "trajectory file or a session folder."
        ),
        type=Path,
    )
    args = argparser.parse_args()
    for path in args.session_paths:
        try:
            inspect_clusters(
                path, images_per_id=args.images_per_id, gt_path=args.gt_path
            )
        except Exception as exc:
            manage_exception(exc)


if __name__ == "__main__":
    idtrackerai_inspect_clusters_entrypoint()



from collections.abc import Iterable, Sequence
from itertools import pairwise

import numpy as np
from qtpy.QtCore import QPointF, QRect, QRectF, Qt
from qtpy.QtGui import QColor, QColorConstants, QImage, QPainter

from idtrackerai import Blob
from idtrackerai.GUI_tools import CanvasPainter


def find_selected_blob(
    blobs_in_frame: list[Blob],
    selected_id: int | None,
    last_position: tuple[float, float] | None,
) -> tuple[Blob | None, tuple[float, float] | None]:
    selected_blobs: list[tuple[Blob, tuple[float, float]]] = []
    for blob in blobs_in_frame:
        for identity, centroid in blob.final_ids_and_centroids:
            if identity == selected_id:
                selected_blobs.append((blob, centroid))

    if len(selected_blobs) == 0:
        return None, last_position
    if len(selected_blobs) == 1:
        return selected_blobs[0]

    if last_position is None:
        return selected_blobs[0]
    prev_cx, prev_cy = last_position
    return min(
        selected_blobs,
        key=lambda blob: (blob[1][0] - prev_cx) ** 2 + (blob[1][1] - prev_cy) ** 2,
    )


def paintBlobs(
    draw_contours: bool,
    draw_centroids: bool,
    draw_bboxes: bool,
    draw_labels: bool,
    painter: CanvasPainter,
    blobs_in_frame: list[Blob],
    cmap: Sequence[QColor],
    cmap_alpha: Sequence[QColor],
    selected_blob: Blob | None,
    selected_centroid: tuple[float, float] | None,
    labels: list[str],
    marked_blobs: Iterable[Blob],
):
    labels_to_draw: list[tuple[QColor, str, tuple]] = []
    painter_window = painter.window()

    if selected_blob is not None:
        selected_blob_final_identities = list(selected_blob.final_identities)
        color_indx = (
            selected_blob_final_identities[0]
            if len(selected_blob_final_identities) == 1
            and selected_blob_final_identities[0] is not None
            else 0
        )
        color_alpha = cmap_alpha[color_indx]

        painter.setPen(QColorConstants.White)
        painter.setBrush(color_alpha)
        painter.drawPolygonFromVertices(selected_blob.contour)
        painter.setBrush(Qt.BrushStyle.NoBrush)

    painter.setPen(Qt.PenStyle.NoPen)
    painter.setBrush(QColor(255, 0, 0, 128))
    for blob in marked_blobs:
        painter.drawPolygonFromVertices(blob.contour)

    for blob in blobs_in_frame:
        x0, y0, x1, y1 = blob.bbox_corners
        if not painter_window.intersects(QRect(x0, y0, x1 - x0, y1 - y0)):
            continue

        blob_final_identities = list(blob.final_identities)
        color_indx = (
            blob_final_identities[0]
            if len(blob_final_identities) == 1 and blob_final_identities[0] is not None
            else 0
        )
        color = cmap[color_indx]

        painter.setPen(color)

        if draw_contours:
            painter.setBrush(Qt.BrushStyle.NoBrush)
            painter.drawPolygonFromVertices(blob.contour)

        if draw_bboxes:
            painter.setBrush(Qt.BrushStyle.NoBrush)
            bbox = blob.bbox_corners
            painter.drawPolygonFromVertices(
                (
                    (bbox.bottom, bbox.left),
                    (bbox.top, bbox.left),
                    (bbox.top, bbox.right),
                    (bbox.bottom, bbox.right),
                )
            )

        for identity, centroid in blob.final_ids_and_centroids:
            if not painter_window.contains(int(centroid[0]), int(centroid[1])):
                continue
            if identity in (None, 0):
                idstr = ""
            else:
                if (
                    blob.user_generated_identities is not None
                    and identity in blob.user_generated_identities
                    and blob.user_generated_centroids is not None
                    and centroid in blob.user_generated_centroids
                ):
                    idstr = "u-" + labels[identity]

                elif (
                    blob.identities_corrected_closing_gaps is not None
                    and not blob.is_an_individual
                ):
                    idstr = "c-" + labels[identity]
                else:
                    idstr = labels[identity]

            color = cmap[0 if identity is None else identity]
            labels_to_draw.append((color, idstr, centroid))

    if selected_blob is not None and selected_centroid is not None:
        radius = 15 * painter.applied_zoom
        x, y = selected_centroid
        painter.setPen(QColorConstants.Black)
        painter.setBrush(Qt.BrushStyle.NoBrush)
        painter.drawEllipse(QRectF(x - radius / 2, y - radius / 2, radius, radius))

    # colored centroids
    if draw_centroids:
        painter.setPen(Qt.PenStyle.NoPen)
        for color, _idstr, (x, y) in labels_to_draw:
            painter.setBrush(color)
            painter.drawBigPoint(x, y)

    # labels lines
    if draw_labels:
        zoom = painter.applied_zoom
        for color, idstr, (x, y) in labels_to_draw:
            if idstr:
                color.setAlpha(90)
                painter.setPen(color)
                painter.drawLine(QPointF(x + 25 * zoom, y - 25 * zoom), QPointF(x, y))

    # black centroid contour
    if draw_centroids:
        painter.setBrush(Qt.BrushStyle.NoBrush)
        painter.setPen(QColorConstants.Black)
        for _color, _idstr, (x, y) in labels_to_draw:
            painter.drawBigPoint(x, y)

    # label text
    if draw_labels:
        zoom = painter.applied_zoom
        for color, idstr, (x, y) in labels_to_draw:
            if idstr:
                color.setAlpha(255)
                painter.setPen(color)
                painter.drawText(QPointF(x + 25 * zoom, y - 25 * zoom), idstr)


def paintTrails(
    frame_number: int,
    painter: CanvasPainter,
    trajectories: np.ndarray,
    cmap: Sequence[QColor],
):
    trail_length = 30
    trail_origin = None if frame_number < trail_length else frame_number - trail_length
    trail_points = trajectories[frame_number:trail_origin:-1]

    if np.isnan(trail_points).all():
        return

    min_x, min_y = np.nanmin(trail_points, axis=(0, 1)).astype(int)
    max_x, max_y = np.nanmax(trail_points, axis=(0, 1)).astype(int)
    traces_bbox = QRect(min_x, min_y, max_x - min_x, max_y - min_y)
    drawing_area = painter.window().intersected(traces_bbox)

    if drawing_area.isEmpty():
        return

    canvas = QImage(
        int(drawing_area.width() / painter.applied_zoom),
        int(drawing_area.height() / painter.applied_zoom),
        QImage.Format.Format_ARGB32_Premultiplied,
    )
    canvas.fill(Qt.GlobalColor.transparent)
    trail_painter = QPainter(canvas)
    trail_painter.setWindow(drawing_area)

    trail_painter.setRenderHint(QPainter.RenderHint.Antialiasing, True)
    trail_painter.setCompositionMode(QPainter.CompositionMode.CompositionMode_Source)

    x0 = drawing_area.x()
    y0 = drawing_area.y()
    x1 = x0 + drawing_area.width()
    y1 = y0 + drawing_area.height()

    pen = trail_painter.pen()
    pen.setWidthF(2 * painter.applied_zoom)
    alphas = np.linspace(255, 0, trail_length, False, dtype=np.uint8)

    are_inside = np.any(
        (trail_points[:, :, 0] > x0)
        * (trail_points[:, :, 0] < x1)
        * (trail_points[:, :, 1] > y0)
        * (trail_points[:, :, 1] < y1),
        0,
    )

    for id_trail_points, is_inside, color in zip(
        trail_points.swapaxes(0, 1), are_inside, cmap[1:]
    ):
        if not is_inside:
            continue

        color = QColor(color)
        for alpha, (A, B) in zip(
            alphas, pairwise(QPointF(*xy) for xy in id_trail_points)
        ):
            color.setAlpha(alpha)
            pen.setColor(color)
            trail_painter.setPen(pen)
            trail_painter.drawLine(A, B)
    trail_painter.end()
    painter.drawImage(drawing_area, canvas)



import json
import logging
from pathlib import Path
from shutil import copyfile

import numpy as np
import torch
from torch.nn import CrossEntropyLoss
from torch.optim.lr_scheduler import MultiStepLR
from torch.optim.sgd import SGD

from idtrackerai import Session, conf
from idtrackerai.utils import load_id_images

from ..network import (
    DEVICE,
    DataLoaderWithLabels,
    IdCNN,
    IdentifierIdCNN,
    StopTraining,
    evaluate_only_acc,
    get_dataloader,
    get_onthefly_dataloader,
    train_loop,
)
from .accumulation_manager import (
    AccumulationManager,
    get_predictions_of_candidates_fragments,
)


def accumulation_step(
    accumulation_manager: AccumulationManager,
    session: Session,
    identification_model: IdCNN,
    model_path: Path,
    penultimate_model_path: Path,
) -> bool:
    logging.info(
        f"[bold]Performing new accumulation, step {accumulation_manager.current_step}",
        extra={"markup": True},
    )
    id_img_paths = session.id_images_file_paths

    # Get images for training
    accumulation_manager.get_new_images_and_labels()

    # get a mixture of old and new images to train
    images_for_training, labels_for_training = (
        accumulation_manager.get_old_and_new_images()
    )

    train_dataloader, val_dataloader, weights = split_data_and_get_dataloaders(
        load_id_images(id_img_paths, images_for_training),
        labels_for_training,
        session.n_animals,
    )

    optimizer = SGD(identification_model.parameters(), lr=0.005, momentum=0.9)

    stopping = StopTraining(
        epochs_limit=conf.MAXIMUM_NUMBER_OF_EPOCHS_IDCNN,
        overfitting_limit=(
            conf.OVERFITTING_COUNTER_THRESHOLD_IDCNN_FIRST_ACCUM
            if accumulation_manager.current_step == 0
            else conf.OVERFITTING_COUNTER_THRESHOLD_IDCNN
        ),
        plateau_limit=conf.LEARNING_RATIO_DIFFERENCE_IDCNN,
    )

    criterion = CrossEntropyLoss(weights).to(DEVICE)
    if conf.TORCH_COMPILE:
        criterion.compile()

    train_loop(
        model=identification_model,
        criterion=criterion,
        optimizer=optimizer,
        train_loader=train_dataloader,
        val_loader=val_dataloader,
        stop_training=stopping,
        scheduler=MultiStepLR(optimizer, milestones=[30, 60], gamma=0.1),
    )

    accumulation_manager.update_fragments_used_for_training()
    accumulation_manager.update_used_images_and_labels()
    accumulation_manager.assign_identities_to_fragments_used_for_training()

    # compute ratio of accumulated images and stop if it is above random
    accumulation_manager.ratio_accumulated_images = (
        accumulation_manager.list_of_fragments.ratio_of_images_used_for_training
    )

    test_acc = test_model(accumulation_manager, id_img_paths, identification_model)

    # keep a copy of the penultimate model
    penultimate_model_path.unlink(missing_ok=True)
    if model_path.is_file():
        copyfile(model_path, penultimate_model_path)
        copyfile(
            model_path.with_suffix(".metadata.json"),
            penultimate_model_path.with_suffix(".metadata.json"),
        )

    logging.info("Saving model at %s", model_path)
    torch.save(identification_model.state_dict(), model_path)
    model_path.with_suffix(".metadata.json").write_text(
        json.dumps(
            {
                "test_acc": test_acc,
                "ratio_accumulated": accumulation_manager.ratio_accumulated_images,
            },
            indent=4,
        )
    )

    if (
        accumulation_manager.ratio_accumulated_images
        > conf.THRESHOLD_EARLY_STOP_ACCUMULATION
    ):
        logging.info(
            "The ratio of accumulated images is higher than"
            f" {conf.THRESHOLD_EARLY_STOP_ACCUMULATION:.1%}, [bold]stopping"
            " accumulation by early stopping criteria",
            extra={"markup": True},
        )
        return True

    # Set accumulation parameters for rest of the accumulation
    # take images from global fragments not used in training (in the remainder test global fragments)
    if any(
        not global_fragment.used_for_training
        for global_fragment in accumulation_manager.list_of_global_fragments
    ):
        logging.info(
            "Generating [bold]predictions[/bold] on remaining global fragments",
            extra={"markup": True},
        )
        (
            predictions,
            probabilities,
            indices_to_split,
            candidate_fragments_identifiers,
        ) = get_predictions_of_candidates_fragments(
            IdentifierIdCNN(identification_model),
            id_img_paths,
            accumulation_manager.list_of_fragments,
        )

        accumulation_manager.split_predictions_after_network_assignment(
            predictions,
            probabilities,
            indices_to_split,
            candidate_fragments_identifiers,
        )

        accumulation_manager.assign_identities()
        accumulation_manager.update_accumulation_statistics()
        accumulation_manager.current_step += 1

    accumulation_manager.ratio_accumulated_images = (
        accumulation_manager.list_of_fragments.ratio_of_images_used_for_training
    )

    session.accumulation_statistics_data = accumulation_manager.accumulation_statistics
    return False


def test_model(
    accumulation_manager: AccumulationManager, id_img_paths: list[Path], model: IdCNN
):
    """Takes a sample of the accumulated images to test model's accuracy"""
    logging.info(
        "Using a sample of all accumulated images to test model's overall accuracy"
    )
    test_images, test_labels = accumulation_manager.get_old_images()
    test_dataloader = get_onthefly_dataloader(test_images, id_img_paths, test_labels)
    test_acc = evaluate_only_acc(test_dataloader, model)
    logging.info(f"Current model has an overall accuracy of {test_acc:.3%}")
    return test_acc


def split_data_and_get_dataloaders(
    images: np.ndarray,
    labels: np.ndarray,
    n_animals: int,
    validation_proportion: float = conf.VALIDATION_PROPORTION,
) -> tuple[DataLoaderWithLabels, DataLoaderWithLabels, torch.Tensor]:
    """Splits a set of `images` and `labels` into training and validation Dataloaders

    Parameters
    ----------
    images : np.ndarray
        List of images (arrays of shape [height, width])
    labels : np.ndarray
        List of integers from 0 to `number_of_animals` - 1
    validation_proportion : float
        The proportion of images that will be used to create the validation set.

    Returns
    -------
    train_dataloader : DataLoaderWithLabels
    val_dataloader : DataLoaderWithLabels
    weights : torch.Tensor
    """
    assert len(images) == len(labels)

    # shuffle images and labels
    shuffled_order = np.arange(len(images))
    np.random.shuffle(shuffled_order)
    images = images[shuffled_order]
    labels = labels[shuffled_order]

    # Init variables
    train_images = []
    train_labels = []
    validation_images = []
    validation_labels = []

    for label in np.unique(labels):
        indices = labels == label
        images_per_class = images[indices]
        labels_per_class = labels[indices]
        n_images_validation = int(validation_proportion * len(labels_per_class) + 0.99)

        validation_images.append(images_per_class[:n_images_validation])
        validation_labels.append(labels_per_class[:n_images_validation])
        train_images.append(images_per_class[n_images_validation:])
        train_labels.append(labels_per_class[n_images_validation:])

    train_images = np.concatenate(train_images)
    train_labels = np.concatenate(train_labels)
    validation_images = np.concatenate(validation_images)
    validation_labels = np.concatenate(validation_labels)

    assert len(train_images) == len(train_labels)
    assert len(validation_images) > 0

    train_dataloader = get_dataloader(
        "training", train_images, train_labels, conf.BATCH_SIZE_IDCNN
    )
    val_dataloader = get_dataloader("validation", validation_images, validation_labels)

    weights = 1 - torch.bincount(
        torch.from_numpy(train_labels), minlength=n_animals
    ).type(torch.float32) / len(train_labels)

    return train_dataloader, val_dataloader, weights



import logging
from collections.abc import Sequence
from pathlib import Path

import numpy as np
from scipy.optimize import linear_sum_assignment
from sklearn.metrics import confusion_matrix

from idtrackerai import GlobalFragment, Session

from ..network import IdentifierBase, get_predictions, load_identifier_model


def identify_first_global_fragment_for_accumulation(
    initial_glob_frag: GlobalFragment,
    session: Session,
    knowledge_transfer_folder: Path | None = None,
    image_size: Sequence[int] | None = None,
):
    logging.info(
        "Using the Global Fragment at frame %d as the first one in accumulation",
        initial_glob_frag.first_frame_of_the_core,
    )

    if knowledge_transfer_folder:
        logging.info(f"Transferring identities from {knowledge_transfer_folder}")
        try:
            identity_transfer_model = load_identifier_model(
                knowledge_transfer_folder, image_size
            )
            identities = get_transferred_identities(
                initial_glob_frag, session.id_images_file_paths, identity_transfer_model
            )
            logging.info("[green]Identity transfer succeeded", extra={"markup": True})
            session.identity_transfer_succeeded = True
        except Exception as exc:
            logging.error(
                "[red bold]Identity transfer failed[/]: %s", exc, extra={"markup": True}
            )
            identities = np.arange(session.n_animals)
            session.identity_transfer_succeeded = False
    else:
        logging.info(
            "Tracking without identity transfer, assigning random initial identities"
        )
        identities = np.arange(session.n_animals)

    for id, fragment in zip(identities, initial_glob_frag):
        fragment.acceptable_for_training = True
        fragment.temporary_id = int(id)
        frequencies = np.zeros(session.n_animals)
        frequencies[id] = fragment.n_images
        fragment.certainty = 1.0
        fragment.set_P1_from_frequencies(frequencies)


def get_transferred_identities(
    initial_glob_frag: GlobalFragment,
    id_images_file_paths: Sequence[Path],
    identification_model: IdentifierBase,
) -> Sequence[int]:
    """Get the identities of the fragments in the initial global fragment using identity transfer."""

    images = []
    frag_id = []
    for i, fragment in enumerate(initial_glob_frag):
        images += fragment.image_locations
        frag_id += [i] * len(fragment)

    predictions, probabilities = get_predictions(
        identification_model, images, id_images_file_paths
    )

    confusion = confusion_matrix(frag_id, predictions - 1, normalize="true")
    frag_ids, identities = linear_sum_assignment(confusion, maximize=True)

    logging.info(
        f"Assigning identities in the {len(initial_glob_frag.fragments)} "
        "fragments of the initial global fragment from identity transfer:\n    "
        + "\n    ".join(
            f"Fragment #{initial_glob_frag.fragments[i].identifier} "
            f"({len(initial_glob_frag.fragments[i])} images long) assigned "
            f"to identity {j + 1} with {confusion[i, j] / np.sum(confusion[i]):.2%} agreement"
            for i, j in zip(frag_ids, identities)
        )
    )

    return identities
